# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager
import errno
import unittest
from unittest import mock

from storlets.sbus import command as sbus_cmd
from storlets.sbus.client import SBusResponse
from storlets.sbus.client.exceptions import SBusClientSendError

from storlets.agent.daemon_factory.server import SDaemonError, \
    StorletDaemonFactory

from tests.unit import FakeLogger
from tests.unit.agent.common import test_server


class DummyDatagram(object):
    def __init__(self, prms=None):
        self.params = prms or {}


class TestStorletDaemonFactory(unittest.TestCase):
    # Class paths used in mock.patch
    base_path = 'storlets.agent.daemon_factory.server'
    os_environ_path = base_path + '.os.environ'
    os_path_exists_path = base_path + '.os.path.exists'
    os_access_path = base_path + '.os.access'
    os_kill_path = base_path + '.os.kill'
    os_waitpid_path = base_path + '.os.waitpid'
    subprocess_popen_path = base_path + '.subprocess.Popen'
    time_sleep_path = base_path + '.time.sleep'

    @contextmanager
    def _mock_sbus_client(self, method):
        sbusclient_path = self.base_path + '.SBusClient'
        with mock.patch('.'.join([sbusclient_path, method])) as _method:
            yield _method

    def setUp(self):
        self.logger = FakeLogger()
        self.pipe_path = 'path/to/pipe'
        self.container_id = 'contid'
        self.dfactory = StorletDaemonFactory(self.pipe_path, self.logger,
                                             self.container_id)

    def test_get_jvm_args(self):
        dummy_env = {'CLASSPATH': '/default/classpath',
                     'LD_LIBRARY_PATH': '/default/ld/library/path'}
        with mock.patch(self.os_environ_path, dummy_env):
            pargs, env = self.dfactory.get_jvm_args(
                'java', 'path/to/storlet/a', 'Storlet-1.0.jar',
                1, 'path/to/uds/a', 'DEBUG')
            self.assertEqual(
                ['/usr/bin/java', 'org.openstack.storlet.daemon.SDaemon',
                 'Storlet-1.0.jar', 'path/to/uds/a', 'DEBUG', '1',
                 self.container_id],
                pargs)

            self.assertIn('CLASSPATH', env)
            self.assertEqual(
                ['/default/classpath',
                 '/usr/local/lib/storlets/java/*',
                 '/usr/local/lib/storlets/java/',
                 'path/to/storlet/a'],
                env['CLASSPATH'].split(':'))

            self.assertIn('LD_LIBRARY_PATH', env)
            self.assertEqual(
                ['/default/ld/library/path',
                 '/usr/local/lib/storlets',
                 '/usr/local/lib/storlets/java'],
                env['LD_LIBRARY_PATH'].split(':'))

    def test_get_python_args(self):
        self._test_get_python_args(None, '3')
        self._test_get_python_args('3.10', '3.10')
        self._test_get_python_args('3', '3')

    def _test_get_python_args(self, version, expected):
        dummy_env = {'PYTHONPATH': '/default/pythonpath'}
        with mock.patch(self.os_environ_path, dummy_env):
            pargs, env = self.dfactory.get_python_args(
                'python', 'path/to/storlet', 'test_storlet.TestStorlet',
                1, 'path/to/uds', 'DEBUG', version)
        self.assertEqual(
            ['/usr/bin/python%s' % expected,
             '/usr/local/libexec/storlets/storlets-daemon',
             'test_storlet.TestStorlet',
             'path/to/uds', 'DEBUG', '1', self.container_id],
            pargs)
        self.assertEqual(
            {'PYTHONPATH': '/default/pythonpath:'
                           '/home/swift/test_storlet.TestStorlet'},
            env)

    def test_spawn_subprocess(self):
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a'}

        class FakePopenObject(object):
            def __init__(self, pid):
                self.pid = pid
                self.stderr = mock.MagicMock()

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen, \
                mock.patch(self.time_sleep_path), \
                mock.patch(self.os_waitpid_path) as waitpid, \
                self._mock_sbus_client('ping') as ping:
            exists.return_value = True
            access.return_velue = True
            popen.side_effect = [FakePopenObject(1000),
                                 FakePopenObject(1001)]
            waitpid.return_value = 0, 0
            ping.return_value = SBusResponse(True, 'OK')
            self.dfactory.spawn_subprocess(
                ['arg0', 'argv1', 'argv2'],
                {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])
            self.assertEqual((1000, 1), waitpid.call_args[0])
            self.assertEqual({'storleta': 1000},
                             self.dfactory.storlet_name_to_pid)

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen, \
                mock.patch(self.time_sleep_path), \
                mock.patch(self.os_waitpid_path) as waitpid, \
                self._mock_sbus_client('ping') as ping:
            exists.return_value = True
            access.return_velue = True
            popen.side_effect = [FakePopenObject(1000),
                                 FakePopenObject(1001)]
            waitpid.return_value = 0, 0
            ping.return_value = SBusResponse(False, 'NG')
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess(
                    ['arg0', 'argv1', 'argv2'],
                    {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])
            self.assertEqual((1000, 1), waitpid.call_args[0])
            self.assertEqual({'storleta': 1000},
                             self.dfactory.storlet_name_to_pid)

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen, \
                mock.patch(self.time_sleep_path), \
                mock.patch(self.os_waitpid_path) as waitpid:
            exists.return_value = True
            access.return_velue = True
            popen.side_effect = [FakePopenObject(1000),
                                 FakePopenObject(1001)]
            waitpid.return_value = 1000, -1
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess(
                    ['arg0', 'argv1', 'argv2'],
                    {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])
            self.assertEqual((1000, 1), waitpid.call_args[0])

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.subprocess_popen_path) as popen:
            exists.return_value = False
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess(
                    ['arg0', 'argv1', 'argv2'],
                    {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])
            popen.assert_not_called()

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen:
            exists.return_value = True
            access.return_value = False
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess(
                    ['arg0', 'argv1', 'argv2'],
                    {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])
            popen.assert_not_called()

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen:
            exists.return_value = True
            access.return_value = False
            popen.side_effect = OSError()
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess(
                    ['arg0', 'argv1', 'argv2'],
                    {'envk0': 'envv0'}, 'storleta')
            self.assertEqual(('arg0',), exists.call_args[0])

        with mock.patch(self.os_path_exists_path) as exists:
            with self.assertRaises(SDaemonError):
                self.dfactory.spawn_subprocess([], {}, 'storleta')
            self.assertEqual(0, exists.call_count)

    def test_wait_for_daemon_to_initialize(self):
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a'}

        with self._mock_sbus_client('ping') as ping, \
                mock.patch(self.time_sleep_path):
            ping.return_value = SBusResponse(True, 'OK')
            self.assertTrue(
                self.dfactory.wait_for_daemon_to_initialize('storleta'))
            self.assertEqual(1, ping.call_count)

        with self._mock_sbus_client('ping') as ping, \
                mock.patch(self.time_sleep_path):
            ping.return_value = SBusResponse(False, 'NG')
            self.assertFalse(
                self.dfactory.wait_for_daemon_to_initialize('storleta'))
            self.assertEqual(
                self.dfactory.NUM_OF_TRIES_PINGING_STARTING_DAEMON,
                ping.call_count)

        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('ping') as ping, \
                mock.patch(self.time_sleep_path):
            ping.side_effect = SBusClientSendError()
            self.assertFalse(
                self.dfactory.wait_for_daemon_to_initialize('storleta'))
            self.assertEqual(
                self.dfactory.NUM_OF_TRIES_PINGING_STARTING_DAEMON,
                ping.call_count)

    def test_process_start_daemon(self):
        # Not running
        self.dfactory.storlet_name_to_pid = {}
        self.dfactory.storlet_name_to_pipe_name = {}

        class FakePopenObject(object):
            def __init__(self, pid):
                self.pid = pid
                self.stderr = mock.MagicMock()

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen, \
                mock.patch(self.time_sleep_path), \
                mock.patch(self.os_waitpid_path) as waitpid, \
                self._mock_sbus_client('ping') as ping:
            exists.return_value = True
            access.return_value = True
            popen.side_effect = [FakePopenObject(1000),
                                 FakePopenObject(1001)]
            waitpid.return_value = 0, 0
            ping.return_value = SBusResponse(True, 'OK')
            self.assertTrue(self.dfactory.process_start_daemon(
                'java', 'path/to/storlet/a', 'storleta', 1, 'path/to/uds/a',
                'TRACE'))
            self.assertEqual({'storleta': 'path/to/uds/a'},
                             self.dfactory.storlet_name_to_pipe_name)

        # Already running
        self.dfactory.storlet_name_to_pid = {'storleta': 1000}
        self.dfactory.storlet_name_to_pipe_name = {'storleta': 'path/to/uds/a'}
        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 0, 0
            self.assertFalse(self.dfactory.process_start_daemon(
                'java', 'path/to/storlet/a', 'storleta', 1, 'path/to/uds/a',
                'TRACE'))

        # Unsupported language
        with self.assertRaises(SDaemonError):
            self.dfactory.process_start_daemon(
                'foo', 'path/to/storlet/a', 'storleta', 1, 'path/to/uds/a',
                'TRACE')

    def test_get_process_status_by_name(self):
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 0, 0
            self.assertTrue(
                self.dfactory.get_process_status_by_name('storleta'))
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

        self.assertFalse(
            self.dfactory.get_process_status_by_name('storletc'))

    def test_get_process_status_by_pid(self):
        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 0, 0
            self.assertTrue(
                self.dfactory.get_process_status_by_pid(1000, 'storleta'))
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 1000, 0
            self.assertFalse(
                self.dfactory.get_process_status_by_pid(1000, 'storleta'))
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = OSError(errno.ESRCH, '')
            self.assertFalse(
                self.dfactory.get_process_status_by_pid(1000, 'storleta'))
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = OSError(errno.EPERM, '')
            exc_pattern = ('^No permission to access the storlet daemon'
                           ' storleta$')
            with self.assertRaisesRegex(SDaemonError, exc_pattern):
                self.dfactory.get_process_status_by_pid(1000, 'storleta')
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = OSError()
            exc_pattern = '^Unknown error$'
            with self.assertRaisesRegex(SDaemonError, exc_pattern):
                self.dfactory.get_process_status_by_pid(1000, 'storleta')
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual((1000, 1), waitpid.call_args[0])

    def test_process_kill(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 1000, 0
            self.assertEqual((1000, 0),
                             self.dfactory.process_kill('storleta'))
            self.assertEqual(1, kill.call_count)
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual({'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # When failed to send kill to the storlet daemon
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            kill.side_effect = OSError()
            with self.assertRaises(SDaemonError):
                self.dfactory.process_kill('storleta')
            self.assertEqual(1, kill.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # When failed to wait
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = OSError()
            with self.assertRaises(SDaemonError):
                self.dfactory.process_kill('storleta')
            self.assertEqual(1, kill.call_count)
            self.assertEqual(1, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # if the storlet daemon is not recognised
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            with self.assertRaises(SDaemonError):
                self.dfactory.process_kill('storletc')
            self.assertEqual(0, kill.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

    def test_process_kill_all(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = [(1000, 0), (1001, 0)]
            self.dfactory.process_kill_all()
            self.assertEqual(2, kill.call_count)
            self.assertEqual(2, waitpid.call_count)
            self.assertEqual({}, self.dfactory.storlet_name_to_pid)

        # Success (no processes)
        self.dfactory.storlet_name_to_pid = {}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            self.dfactory.process_kill_all()
            self.assertEqual(0, kill.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({}, self.dfactory.storlet_name_to_pid)

        # Failure (try_all = True)
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            kill.side_effect = OSError()
            exc_pattern = '^Failed to kill some storlet daemons: .*'
            with self.assertRaisesRegex(SDaemonError, exc_pattern) as e:
                self.dfactory.process_kill_all()
            self.assertIn('storleta', str(e.exception))
            self.assertIn('storletb', str(e.exception))
            self.assertEqual(2, kill.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # Failure (try_all = False)
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path) as waitpid:
            kill.side_effect = OSError()
            exc_pattern = ('^Failed to send kill signal to the storlet daemon '
                           'storlet[a-b]$')
            with self.assertRaisesRegex(SDaemonError, exc_pattern):
                self.dfactory.process_kill_all(False)
            self.assertEqual(1, kill.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

    def test_shutdown_all_processes(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path):
            halt.return_value = SBusResponse(True, 'OK')
            terminated = self.dfactory.shutdown_all_processes()
            self.assertEqual(2, len(terminated))
            self.assertIn('storleta', terminated)
            self.assertIn('storletb', terminated)
            self.assertEqual({},
                             self.dfactory.storlet_name_to_pid)

        # Failure (try_all = True)
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'patha', 'storletb': 'pathb'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path) as waitpid:
            halt.side_effect = SBusClientSendError()
            exc_pattern = '^Failed to shutdown some storlet daemons: .*'
            with self.assertRaisesRegex(SDaemonError, exc_pattern) as e:
                self.dfactory.shutdown_all_processes()
            self.assertIn('storleta', str(e.exception))
            self.assertIn('storletb', str(e.exception))
            self.assertEqual(2, halt.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # Failure (try_all = False)
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'patha', 'storletb': 'pathb'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path) as waitpid:
            halt.side_effect = SBusClientSendError()
            exc_pattern = ('^Failed to send halt command to the storlet '
                           'daemon storlet[a-b]$')
            with self.assertRaisesRegex(SDaemonError, exc_pattern):
                self.dfactory.shutdown_all_processes(False)
            self.assertEqual(1, halt.call_count)
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

    def test_shutdown_process(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path):
            halt.return_value = SBusResponse(True, 'OK')
            self.dfactory.shutdown_process('storleta')
            self.assertEqual({'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # Failed to send a command to the storlet daemon
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path) as waitpid:
            halt.side_effect = SBusClientSendError()
            with self.assertRaises(SDaemonError):
                self.dfactory.shutdown_process('storleta')
            self.assertEqual(0, waitpid.call_count)
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # Failed to wait
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path) as waitpid:
            halt.return_value = SBusResponse(True, 'OK')
            waitpid.side_effect = OSError()
            with self.assertRaises(SDaemonError):
                self.dfactory.shutdown_process('storleta')
            self.assertEqual({'storleta': 1000, 'storletb': 1001},
                             self.dfactory.storlet_name_to_pid)

        # If the storlet is not found in pid mapping
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self.assertRaises(SDaemonError):
            self.dfactory.shutdown_process('storletc')

    def test_start_daemon(self):
        prms = {'daemon_language': 'java',
                'storlet_path': 'path/to/storlet/a',
                'storlet_name': 'storleta',
                'pool_size': 1,
                'uds_path': 'path/to/uds/a',
                'log_level': 'TRACE'}
        # Not running
        self.dfactory.storlet_name_to_pid = {}
        self.dfactory.storlet_name_to_pipe_name = {}

        class FakePopenObject(object):
            def __init__(self, pid):
                self.pid = pid
                self.stderr = mock.MagicMock()

        with mock.patch(self.os_path_exists_path) as exists, \
                mock.patch(self.os_access_path) as access, \
                mock.patch(self.subprocess_popen_path) as popen, \
                mock.patch(self.time_sleep_path), \
                mock.patch(self.os_waitpid_path) as waitpid, \
                self._mock_sbus_client('ping') as ping, \
                self._mock_sbus_client('start_daemon') as start_daemon:
            exists.return_value = True
            access.return_value = True
            popen.side_effect = [FakePopenObject(1000),
                                 FakePopenObject(1001)]
            waitpid.return_value = 0, 0
            ping.return_value = SBusResponse(True, 'OK')
            start_daemon.return_value = SBusResponse(True, 'OK')
            ret = self.dfactory.start_daemon(DummyDatagram(prms))
            self.assertTrue(ret.status)
            self.assertEqual('OK', ret.message)
            self.assertTrue(ret.iterable)

        # Already running
        self.dfactory.storlet_name_to_pid = {'storleta': 1000}
        self.dfactory.storlet_name_to_pipe_name = {'storleta': 'path/to/uds/a'}
        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 0, 0
            ret = self.dfactory.start_daemon(DummyDatagram(prms))
            self.assertTrue(ret.status)
            self.assertEqual(
                'The storlet daemon storleta is already running', ret.message)
            self.assertTrue(ret.iterable)

        # Unsupported language
        prms['daemon_language'] = 'foo'
        ret = self.dfactory.start_daemon(DummyDatagram(prms))
        self.assertFalse(ret.status)
        self.assertEqual('Got unsupported daemon language: foo', ret.message)
        self.assertTrue(ret.iterable)

    def test_stop_daemon(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000}
        with mock.patch(self.os_kill_path), \
                mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 1000, 0
            resp = self.dfactory.stop_daemon(
                DummyDatagram({'storlet_name': 'storleta'}))
            self.assertTrue(resp.status)
            self.assertEqual('The storlet daemon storleta is stopped',
                             resp.message)
            self.assertTrue(resp.iterable)

        # Failure
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000}
        with mock.patch(self.os_kill_path) as kill, \
                mock.patch(self.os_waitpid_path):
            kill.side_effect = OSError('ERROR')
            resp = self.dfactory.stop_daemon(
                DummyDatagram({'storlet_name': 'storleta'}))
            self.assertFalse(resp.status)
            self.assertEqual(
                'Failed to send kill signal to the storlet daemon storleta',
                resp.message)
            self.assertTrue(resp.iterable)

    def test_daemon_status(self):
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 0, 0
            resp = self.dfactory.daemon_status(
                DummyDatagram({'storlet_name': 'storleta'}))
            self.assertTrue(resp.status)
            self.assertEqual('The storlet daemon storleta seems to be OK',
                             resp.message)
            self.assertTrue(resp.iterable)

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.return_value = 1000, 0
            resp = self.dfactory.daemon_status(
                DummyDatagram({'storlet_name': 'storleta'}))
            self.assertFalse(resp.status)
            self.assertEqual('No running storlet daemons for storleta',
                             resp.message)
            self.assertTrue(resp.iterable)

        with mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = OSError()
            resp = self.dfactory.daemon_status(
                DummyDatagram({'storlet_name': 'storleta'}))
            self.assertFalse(resp.status)
            self.assertEqual('Unknown error', resp.message)
            self.assertTrue(resp.iterable)

    def test_halt(self):
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        self.dfactory.storlet_name_to_pipe_name = \
            {'storleta': 'path/to/uds/a', 'storletb': 'path/to/uds/b'}
        with self._mock_sbus_client('halt') as halt, \
                mock.patch(self.os_waitpid_path):
            halt.return_value = SBusResponse(True, 'OK')
            resp = self.dfactory.halt(DummyDatagram())
            self.assertTrue(resp.status)
            self.assertEqual(
                'Stopped all storlet daemons. Terminating.', resp.message)
            self.assertFalse(resp.iterable)

    def test_stop_daemons(self):
        # Success
        self.dfactory.storlet_name_to_pid = \
            {'storleta': 1000, 'storletb': 1001}
        with mock.patch(self.os_kill_path), \
                mock.patch(self.os_waitpid_path) as waitpid:
            waitpid.side_effect = [(1000, 0), (1001, 0)]
            resp = self.dfactory.stop_daemons(DummyDatagram())
            self.assertTrue(resp.status)
            self.assertEqual('OK', resp.message)
            self.assertFalse(resp.iterable)


class TestSBusServerMain(test_server.TestSBusServerMain):

    def _get_test_server(self):
        return StorletDaemonFactory(self.sbus_path, self.logger, 'contid')

    def test_main_loop_successful_stop(self):
        # SBUS_CMD_HALT is for working to stop requested from
        # storlet_middleware
        self._test_main_loop_stop(sbus_cmd.SBUS_CMD_HALT)

    def test_main_loop_timeout(self):
        self._test_main_loop_timeout()


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import storlets.sbus.file_description as sbus_fd
from storlets.sbus.datagram import SBusFileDescriptor, SBusDatagram, \
    SBusServiceDatagram, SBusExecuteDatagram, build_datagram, \
    build_datagram_from_raw_message
from storlets.sbus.command import SBUS_CMD_PING, SBUS_CMD_EXECUTE

ALL_FDTYPES = [
    sbus_fd.SBUS_FD_INPUT_OBJECT, sbus_fd.SBUS_FD_OUTPUT_OBJECT,
    sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
    sbus_fd.SBUS_FD_OUTPUT_OBJECT_AND_METADATA,
    sbus_fd.SBUS_FD_LOGGER, sbus_fd.SBUS_FD_OUTPUT_CONTAINER,
    sbus_fd.SBUS_FD_SERVICE_OUT,
]


class TestSBusFileDescriptor(unittest.TestCase):
    def test_metadata(self):
        fd = SBusFileDescriptor(
            'MYTYPE', 1, {'storlets_key': 'storlets_value'},
            {'storage_key': 'storage_value'})
        self.assertEqual(
            {'storlets': {'type': 'MYTYPE', 'storlets_key': 'storlets_value'},
             'storage': {'storage_key': 'storage_value'}},
            fd.metadata)

    def test_from_fileno_and_metadata_dict(self):
        fd = SBusFileDescriptor.from_fileno_and_metadata_dict(
            1,
            {'storlets': {'type': 'MYTYPE', 'storlets_key': 'storlets_value'},
             'storage': {'storage_key': 'storage_value'}})
        self.assertEqual(1, fd.fileno)
        self.assertEqual('MYTYPE', fd.fdtype)
        self.assertEqual({'storlets_key': 'storlets_value'},
                         fd.storlets_metadata)
        self.assertEqual({'storage_key': 'storage_value'},
                         fd.storage_metadata)


class TestSBusDatagram(unittest.TestCase):
    def test_check_required_fdtypes_not_implemented(self):
        # SBusDatagram designed not to be called independently
        with self.assertRaises(NotImplementedError) as err:
            SBusDatagram('', [], [])
        self.assertEqual(
            'SBusDatagram class should not be initialized as bare',
            err.exception.args[0])

    def test_invalid_child_class_definition(self):
        # no definition for _required_fdtypes
        class InvalidSBusDatagram(SBusDatagram):
            pass

        with self.assertRaises(NotImplementedError) as err:
            InvalidSBusDatagram('', [], [])
        self.assertEqual(
            'SBusDatagram class should define _required_fdtypes',
            err.exception.args[0])


class SBusDatagramTestMixin(object):
    def setUp(self):
        self.params = {'param1': 'paramvalue1'}
        self.task_id = 'id'
        self.dtg = self._test_class(self.command, self.sfds,
                                    self.params, self.task_id)

    def test_init(self):
        self.assertEqual(self.command, self.dtg.command)
        self.assertEqual(self.sfds, self.dtg.sfds)
        self.assertEqual(self.params, self.dtg.params)
        self.assertEqual(self.task_id, self.dtg.task_id)

    def test_num_fds(self):
        self.assertEqual(len(self.fdtypes), self.dtg.num_fds)

    def test_cmd_params(self):
        self.assertEqual({'command': self.command,
                          'params': self.params,
                          'task_id': self.task_id},
                         self.dtg.cmd_params)

    def test_serialized_cmd_params(self):
        res = {'command': self.command,
               'params': self.params,
               'task_id': self.task_id}
        self.assertEqual(res, json.loads(self.dtg.serialized_cmd_params))

    def test_check_required_fdtypes_mismatch(self):
        invalid_fdtypes_list = (
            [],  # empty list
            ['Invalid'] + self.fdtypes,  # invalid type inserted at the first
            # TODO(kota_): we may want *strict* check (not only checking first
            #              N items.
        )

        for invalid_fdtypes in invalid_fdtypes_list:
            with self.assertRaises(ValueError) as cm:
                self.dtg._check_required_fdtypes(invalid_fdtypes)
            self.assertTrue(cm.exception.args[0].startswith(
                'Fd type mismatch given_fdtypes'))

    def test_find_fds(self):
        # prepare all fd types and then pop out in the loop below
        not_in_fdtypes = ALL_FDTYPES[:]
        # N.B. fd should start from 1 (not 0), really?
        for index, fdtype in enumerate(self.fdtypes, 1):
            found_fds = self.dtg._find_fds(fdtype)
            # at least 1 fd should be found
            self.assertTrue(found_fds)
            # and the index is in the types
            self.assertIn(index, found_fds)

            if fdtype in not_in_fdtypes:
                # N.B. ALL_FDTYPES should be unique list
                not_in_fdtypes.remove(fdtype)

        # sanity, not a fd type results in []
        self.assertEqual([], self.dtg._find_fds('DUMMY_TYPE'))

        # sanity, no other types are found
        for fdtype in not_in_fdtypes:
            self.assertEqual([], self.dtg._find_fds(fdtype))

    def test_find_fd(self):
        # prepare all fd types and then pop out in the loop below
        not_in_fdtypes = ALL_FDTYPES[:]
        # N.B. fd should start from 1 (not 0), really?
        for index, fdtype in enumerate(self.fdtypes, 1):
            found_fd = self.dtg._find_fd(fdtype)
            # at least 1 fd should be found
            self.assertEqual(index, found_fd)

            if fdtype in not_in_fdtypes:
                # N.B. ALL_FDTYPES should be unique list
                not_in_fdtypes.remove(fdtype)

        # sanity, not a fd type results in None
        self.assertIsNone(self.dtg._find_fd('DUMMY_TYPE'))

        # sanity, no other types are found
        for fdtype in not_in_fdtypes:
            self.assertIsNone(self.dtg._find_fd(fdtype))


class TestSBusServiceDatagram(SBusDatagramTestMixin, unittest.TestCase):
    _test_class = SBusServiceDatagram

    def setUp(self):
        self.command = 'SBUS_CMD_TEST'
        self.fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT]
        self.sfds = [SBusFileDescriptor(sbus_fd.SBUS_FD_SERVICE_OUT, 1)]
        super(TestSBusServiceDatagram, self).setUp()

    def test_service_out_fd(self):
        self.assertEqual(1, self.dtg.service_out_fd)


class TestSBusExecuteDatagram(SBusDatagramTestMixin, unittest.TestCase):
    _test_class = SBusExecuteDatagram

    def setUp(self):
        self.command = SBUS_CMD_EXECUTE
        self.fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT,
                        sbus_fd.SBUS_FD_INPUT_OBJECT,
                        sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                        sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                        sbus_fd.SBUS_FD_LOGGER]
        self.sfds = [SBusFileDescriptor(
            fdtype, i + 1,
            {'key%d' % i: 'value%d' % i},
            {'skey%d' % i: 'svalue%d' % i})
            for i, fdtype in enumerate(self.fdtypes)]
        super(TestSBusExecuteDatagram, self).setUp()

    def test_init_extra_sources(self):
        fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                   sbus_fd.SBUS_FD_LOGGER,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT]
        fds = [SBusFileDescriptor(fdtype, i + 1,
               {'key%d' % i: 'value%d' % i},
               {'skey%d' % i: 'svalue%d' % i})
               for i, fdtype in enumerate(fdtypes)]
        dtg = self._test_class(
            self.command, fds, self.params, self.task_id)
        self.assertEqual(fdtypes, [sfd.fdtype for sfd in dtg.sfds])
        self.assertEqual(self.params, dtg.params)
        self.assertEqual(self.task_id, dtg.task_id)

    def test_service_out_fd(self):
        self.assertEqual(1, self.dtg.service_out_fd)

    def test_invocation_fds(self):
        self.assertEqual([2, 3, 4, 5], self.dtg.invocation_fds)

    def test_object_out_fds(self):
        self.assertEqual([3], self.dtg.object_out_fds)

    def test_object_metadata_out_fds(self):
        self.assertEqual([4], self.dtg.object_metadata_out_fds)

    def test_logger_out_fd(self):
        self.assertEqual(5, self.dtg.logger_out_fd)

    def test_object_in_fds(self):
        self.assertEqual([2], self.dtg.object_in_fds)

    def test_check_required_fdtypes_reverse_order_failed(self):
        fdtypes = self.fdtypes[:]
        fdtypes.reverse()  # reverse order
        with self.assertRaises(ValueError) as cm:
            self.dtg._check_required_fdtypes(fdtypes)
        self.assertTrue(
            cm.exception.args[0].startswith('Fd type mismatch given_fdtypes'))


class TestBuildDatagram(unittest.TestCase):

    def test_build_datagram(self):
        # SBusServiceDatagram scenario
        command = SBUS_CMD_PING
        fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT]
        fds = [SBusFileDescriptor(sbus_fd.SBUS_FD_SERVICE_OUT, 1)]
        params = {'param1': 'paramvalue1'}
        task_id = 'id'

        dtg = build_datagram(command, fds, params, task_id)

        self.assertIsInstance(dtg, SBusServiceDatagram)
        self.assertEqual(command, dtg.command)
        self.assertEqual(fdtypes, [sfd.fdtype for sfd in dtg.sfds])
        self.assertEqual(params, dtg.params)
        self.assertEqual(task_id, dtg.task_id)

        # SBusExecuteDatagram scenario
        command = SBUS_CMD_EXECUTE
        fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                   sbus_fd.SBUS_FD_LOGGER,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT]
        fds = [SBusFileDescriptor(fdtype, i + 1,
               {'key%d' % i: 'value%d' % i},
               {'skey%d' % i: 'svalue%d' % i})
               for i, fdtype in enumerate(fdtypes)]
        params = {'param1': 'paramvalue1'}
        task_id = 'id'

        dtg = build_datagram(command, fds, params, task_id)

        self.assertIsInstance(dtg, SBusExecuteDatagram)
        self.assertEqual(command, dtg.command)
        self.assertEqual(fdtypes, [sfd.fdtype for sfd in dtg.sfds])
        self.assertEqual(params, dtg.params)
        self.assertEqual(task_id, dtg.task_id)

    def test_build_datagram_from_raw_message(self):
        # SBusServiceDatagram scenario
        command = SBUS_CMD_PING
        fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT]
        fds = [SBusFileDescriptor(sbus_fd.SBUS_FD_SERVICE_OUT, 1)]
        params = {'param1': 'paramvalue1'}
        task_id = 'id'
        cmd_params = {'command': command, 'params': params, 'task_id': task_id}

        str_metadata = json.dumps([fd.metadata for fd in fds])
        str_cmd_params = json.dumps(cmd_params)
        dtg = build_datagram_from_raw_message(fds, str_metadata,
                                              str_cmd_params)

        self.assertIsInstance(dtg, SBusServiceDatagram)
        self.assertEqual(command, dtg.command)
        self.assertEqual(fdtypes, [sfd.fdtype for sfd in dtg.sfds])
        self.assertEqual(params, dtg.params)
        self.assertEqual(task_id, dtg.task_id)

        # SBusExecuteDatagram scenario
        command = SBUS_CMD_EXECUTE
        fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                   sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                   sbus_fd.SBUS_FD_LOGGER,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT,
                   sbus_fd.SBUS_FD_INPUT_OBJECT]
        fds = [SBusFileDescriptor(fdtype, i + 1,
               {'key%d' % i: 'value%d' % i},
               {'skey%d' % i: 'svalue%d' % i})
               for i, fdtype in enumerate(fdtypes)]
        params = {'param1': 'paramvalue1'}
        task_id = 'id'
        cmd_params = {'command': command, 'params': params, 'task_id': task_id}

        str_metadata = json.dumps([fd.metadata for fd in fds])
        str_cmd_params = json.dumps(cmd_params)
        dtg = build_datagram_from_raw_message(fds, str_metadata,
                                              str_cmd_params)

        self.assertIsInstance(dtg, SBusExecuteDatagram)
        self.assertEqual(command, dtg.command)
        self.assertEqual(fdtypes, [sfd.fdtype for sfd in dtg.sfds])
        self.assertEqual(params, dtg.params)
        self.assertEqual(task_id, dtg.task_id)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from swiftclient import client as swift_client
from swiftclient import ClientException
from storlets.tools.utils import get_member_auth
from tests.functional.python import StorletPythonFunctionalTest
import unittest


class myTestThread(threading.Thread):
    def __init__(self, url, token, test_class):
        super(myTestThread, self).__init__()
        self.token = token
        self.url = url
        self.test_class = test_class

    def run(self):
        self.test_class.invoke_storlet("print", False)


class TestTestStorlet(StorletPythonFunctionalTest):
    def setUp(self):
        self.additional_headers = {}
        super(TestTestStorlet, self).setUp(
            storlet_dir='test',
            storlet_name='test.py',
            storlet_main='test.TestStorlet',
            storlet_file=None,
            headers={})

        self.member_url, self.member_token = get_member_auth(self.conf)

        swift_client.put_object(self.url,
                                self.token,
                                self.container,
                                'test_object',
                                'some content')

    def invoke_storlet(self, op, withlog=False):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        if withlog is True:
            headers['X-Storlet-Generate-Log'] = 'True'

        params = 'op={0}&param2=val2'.format(op)
        resp_dict = dict()
        try:
            resp_headers, get_text = swift_client.get_object(
                self.url, self.token, self.container, 'test_object',
                None, None, params, resp_dict, headers)
            get_response_status = resp_dict.get('status')

        except Exception:
            get_response_status = resp_dict.get('status')
            if op == 'crash':
                self.assertTrue(get_response_status >= 500 or
                                get_response_status == 404)

        if withlog:
            resp_headers, get_text = swift_client.get_object(
                self.url, self.token, 'storletlog', 'test.log',
                None, None, None, None, headers)
            self.assertEqual(200, resp_headers.get('status'))
            self.assertEqual(b'aaa', get_text.read())

        if op == 'print':
            self.assertEqual(200, get_response_status)
            self.assertIn(b'op     print', get_text)
            self.assertIn(b'param2     val2', get_text)

    def test_print(self):
        self.invoke_storlet("print", False)

    def test_crash(self):
        self.invoke_storlet("crash")

    def test_hold(self):
        self.invoke_storlet("hold")

    test_hold.slow = 1

    def invoke_storlet_in_parallel(self):
        mythreads = []

        for i in range(10):
            new_thread = myTestThread(self.url, self.token, self)
            mythreads.append(new_thread)

        for t in mythreads:
            t.start()

        for t in mythreads:
            t.join()

    def test_parallel_print(self):
        self.invoke_storlet_in_parallel()

    test_parallel_print.slow = 1

    def test_storlet_acl_get_fail(self):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        exc_pattern = '^.*403 Forbidden.*$'
        with self.assertRaisesRegex(ClientException, exc_pattern):
            swift_client.get_object(self.member_url, self.member_token,
                                    self.container, 'test_object',
                                    headers=headers)

    def test_storlet_acl_get_success(self):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        exc_pattern = '^.*403 Forbidden.*$'
        with self.assertRaisesRegex(ClientException, exc_pattern):
            swift_client.get_object(self.member_url, self.member_token,
                                    self.container, 'test_object',
                                    headers=headers)

        headers = {'X-Storlet-Container-Read': self.conf.member_user,
                   'X-Storlet-Name': self.storlet_name}
        swift_client.post_container(self.url,
                                    self.token,
                                    self.container,
                                    headers)
        swift_client.head_container(self.url,
                                    self.token,
                                    self.container)
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        resp_dict = dict()
        swift_client.get_object(self.member_url,
                                self.member_token,
                                self.container, 'test_object',
                                response_dict=resp_dict,
                                headers=headers)
        self.assertEqual(200, resp_dict['status'])


class TestTestStorletOnProxy(TestTestStorlet):
    def setUp(self):
        super(TestTestStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from io import StringIO


class FakeFileManager(object):
    def __init__(self, storlet, dependency, storlet_perm=None,
                 dependency_perm=None):
        self.storlet = storlet
        self.dependency = dependency
        self.storlet_perm = storlet_perm
        self.dependency_perm = dependency_perm

    def get_storlet(self, name):
        return StringIO(self.storlet), self.storlet_perm

    def get_dependency(self, name):
        return StringIO(self.dependency), self.dependency_perm

    def put_log(self, name, fobj):
        pass



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import traceback
import functools
from collections import defaultdict
from shutil import rmtree
from tempfile import mkdtemp


def with_tempdir(f):
    """
    Decorator to give a single test a tempdir as argument to test method.
    """
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        tempdir = mkdtemp()
        args = list(args)
        args.append(tempdir)
        try:
            return f(*args, **kwargs)
        finally:
            rmtree(tempdir)
    return wrapped


class FakeLogger(object):
    def __init__(self, *args, **kwargs):
        self._log_lines = defaultdict(list)

    def _print_log(self, level, msg):
        self._log_lines[level.lower()].append(msg)
        print('%s: %s' % (level, msg))

    def get_log_lines(self, level):
        return self._log_lines[level.lower()]

    def debug(self, msg):
        self._print_log('DEBUG', msg)

    def info(self, msg):
        self._print_log('INFO', msg)

    def warning(self, msg):
        self._print_log('WARN', msg)

    warn = warning

    def error(self, msg):
        self._print_log('ERROR', msg)

    def exception(self, msg):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        exc_msg = traceback.format_exception(exc_type, exc_value,
                                             exc_traceback)
        new_msg = '%s: %s' % (msg, exc_msg)
        self._print_log('EXCEPTION', new_msg)



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from tests.functional import StorletFunctionalTest, PATH_TO_STORLETS

BIN_DIR = 'bin'


class StorletJavaFunctionalTest(StorletFunctionalTest):
    def setUp(self, storlet_dir, storlet_name, storlet_main,
              storlet_file, dep_names=None, headers=None):
        storlet_dir = os.path.join('java', storlet_dir)
        path_to_bundle = os.path.join(PATH_TO_STORLETS, storlet_dir,
                                      BIN_DIR)
        super(StorletJavaFunctionalTest, self).setUp('Java',
                                                     path_to_bundle,
                                                     storlet_dir,
                                                     storlet_name,
                                                     storlet_main,
                                                     storlet_file,
                                                     dep_names,
                                                     headers)



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
from storlets.tools import deploy_storlet
from tests.functional import StorletBaseFunctionalTest, PATH_TO_STORLETS
from tests.functional.java import BIN_DIR
from tests.functional.common.mixins import DeployTestMixin, EXPECT, SEND_LINE


class TestDeployStorlet(DeployTestMixin, StorletBaseFunctionalTest):
    def setUp(self):
        super(TestDeployStorlet, self).setUp()
        self.deploy_storlet_path = os.path.abspath(deploy_storlet.__file__)
        self.execdep_storlet_path = os.path.join(PATH_TO_STORLETS,
                                                 'java',
                                                 'ExecDepStorlet',
                                                 BIN_DIR)
        self.execdep_storlet_file = os.path.join(self.execdep_storlet_path,
                                                 'execdepstorlet-1.0.jar')
        self.execdep_storlet_dep_file = os.path.join(self.execdep_storlet_path,
                                                     'get42')

        self.scenario = [
            (b'Enter storlet language (java or python): ', EXPECT),
            (b'java', SEND_LINE),
            (b'Enter absolute path to storlet jar file: ', EXPECT),
            (self.execdep_storlet_file.encode('ascii'), SEND_LINE),
            (b'Your jar file contains the following classes:', EXPECT),
            (b'\t* org.openstack.storlet.execdep.ExecDepStorlet', EXPECT),
            (b'Please enter fully qualified storlet main class '
             b'(choose from the list above): ', EXPECT),
            (b'org.openstack.storlet.execdep.ExecDepStorlet', SEND_LINE),
            (b'Please enter dependency files '
             b'(leave a blank line when you are done): ', EXPECT),
            (self.execdep_storlet_dep_file.encode('ascii'), SEND_LINE),
            (b'', SEND_LINE),  # DO NOT send \n but just empty due to sendline
            (b'Storlet deployment complete', EXPECT),
        ]


if __name__ == '__main__':
    unittest.main()



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import zipfile
from storlets.tools.cluster_config_parser import ClusterConfig
from storlets.tools.utils import get_auth, deploy_storlet

CLS_SUFFIX = '.class'


class StdinReader(object):
    """
    Helper reader class to get stdin and map the value to specified key

    :param key: a string but it's never used in this class just a reference
                for the value should be assigned
    :param run_once: bool. If set False, continue to readline until blank line
                     in __call__ method.
    :param callback: function which should be called with the value after
                     all lines were read.
    """
    def __init__(self, key, run_once=True, callback=None):
        self.key = key
        self.run_once = run_once
        self.callback = callback

    def __call__(self):
        """
        :return: string when run_once == True else list
        """
        line = sys.stdin.readline().rstrip()
        if not self.run_once:
            value = []
            while line:
                value.append(line)
                line = sys.stdin.readline().rstrip()
        else:
            value = line

        if self.callback:
            self.callback(value)

        return value


def list_classes(storlet_file):
    print("Your jar file contains the following classes:")
    with zipfile.ZipFile(storlet_file, 'r') as zfile:
        for f in zfile.infolist():
            name = f.filename
            if name.endswith(CLS_SUFFIX):
                print('\t* ' +
                      name[:-len(CLS_SUFFIX)].replace('/', '.'))


MESSAGES = {
    'Java': iter([
        ("Enter absolute path to storlet jar file: ",
         StdinReader("storlet", callback=list_classes)),
        ("Please enter fully qualified storlet main class "
         "(choose from the list above): ", StdinReader("storlet_main_class")),
        ("Please enter dependency files "
         "(leave a blank line when you are done):",
         StdinReader("dependencies", False))]),
    'Python': iter([
        ("Enter absolute path to storlet file: ", StdinReader("storlet")),
        ("Please enter fully qualified storlet main class "
         "<filename.ClassName>: ", StdinReader("storlet_main_class")),
        ("Please enter dependency files "
         "(leave a blank line when you are done): ",
         StdinReader("dependencies", False))]),
}


def usage():
    print("Usage: deploy_storlet.py <path to conf>")


def main(argv):
    if len(argv) != 1:
        usage()
        return 1
    conf = ClusterConfig(argv[0])
    url, token = get_auth(conf, conf.admin_user, conf.admin_password)
    print("Enter storlet language (java or python): ")
    storlet_language = sys.stdin.readline().rstrip().title()

    if storlet_language not in MESSAGES:
        print("The language you specified is not supported")
        return 1

    message_iter = MESSAGES[storlet_language]
    options_dict = dict(language=storlet_language)
    for message, reader in message_iter:
        print(message)
        options_dict[reader.key] = reader()

    deploy_storlet(url, token, **options_dict)

    print("Storlet deployment complete")
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

from storlets.gateway.common.stob import StorletRequest
from storlets.gateway.gateways.base import StorletGatewayBase
from storlets.gateway.gateways.container.runtime import RunTimePaths, \
    StorletInvocationProtocol


class ContainerStorletRequest(StorletRequest):

    required_options = ['storlet_main', 'storlet_language', 'file_manager']

    def __init__(self, storlet_id, params, data, options=None,
                 extra_data_list=None):
        """
        :param storlet_id: storlet id
        :param params: execution parameters
        :param data: StorletData instance
        :param options: options specific to ContainerStorletRequest
        :param extra_data_list: List of StorletData instances
        :raises ValueError: when some of the required options (storlet_main
                            and file_manager) are missing
        """
        super(ContainerStorletRequest, self).__init__(
            storlet_id, params, data, options, extra_data_list)

        self.generate_log = self.options.get('generate_log', False)

        self.storlet_main = self.options['storlet_main']
        self.storlet_language = self.options['storlet_language']
        self.storlet_language_version = \
            self.options.get('storlet_language_version')

        if self.options.get('storlet_dependency'):
            self.dependencies = [
                x.strip() for x
                in self.options['storlet_dependency'].split(',')
                if x.strip()]
        else:
            self.dependencies = []

        self.file_manager = self.options['file_manager']

        self.start = self.options.get('range_start')
        self.end = self.options.get('range_end')

    @property
    def has_range(self):
        """
        Whether the input range is given
        """
        return self.start is not None and self.end is not None


class StorletGatewayContainer(StorletGatewayBase):

    request_class = ContainerStorletRequest
    sandbox = None

    def __init__(self, conf, logger, scope):
        """
        :param conf: a dict for gateway conf
        :param logger: a logger instance
        :param scope: scope name to identify the container
        """
        super(StorletGatewayContainer, self).__init__(conf, logger, scope)
        self.storlet_timeout = float(self.conf.get('storlet_timeout', 40))
        self.paths = RunTimePaths(scope, conf)

    @classmethod
    def validate_storlet_registration(cls, params, name):
        """
        Validate required parameters for storlet file

        :param params: parameters related to the storlet file
        :param name: name of the storlet file
        :raises ValueError: if some of the required parameters are missing,
                            or some of the parameters are invalid
        """
        mandatory = ['Language', 'Interface-Version', 'Object-Metadata',
                     'Main']
        cls._check_mandatory_params(params, mandatory)

        if params['Language'].lower() == 'java':
            if '-' not in name or '.' not in name:
                raise ValueError('Storlet name is incorrect')
        elif params['Language'].lower() == 'python':
            try:
                version = float(params.get('Language-Version', 3))
            except ValueError:
                raise ValueError('Language-Version is invalid')

            if int(version) != 3:
                # TODO(kota_): more strict version check should be nice.
                raise ValueError('Not supported version specified')

            if name.endswith('.py'):
                cls_name = params['Main']
                if not cls_name.startswith(name[:-3] + '.'):
                    raise ValueError('Main class should be included in '
                                     'storlet file')

                if len(cls_name.split('.')) != 2:
                    raise ValueError('Submodule is currently not supported')
            # TODO(takashi): Add support for sdist tar.gz
            else:
                raise ValueError('Storlet name is incorrect')
        else:
            raise ValueError('Unsupported Language')

        dep = params.get('Dependency')
        if dep:
            deps = dep.split(',')
            if name in deps:
                raise ValueError('Using the same name for storlet and '
                                 'dependency is not allowed')
            if len(deps) != len(set(deps)):
                raise ValueError('Duplicated name in dependencies')

    @classmethod
    def validate_dependency_registration(cls, params, name):
        """
        Validate required parameters for dependency file

        :param params: parameters related to the dependency file
        :param name: name of the dependency file
        :raises ValueError: if some of the required parameters are missing,
                            or some of the parameters are invalid
        """
        mandatory = ['Dependency-Version']
        cls._check_mandatory_params(params, mandatory)

        perm = params.get('Dependency-Permissions')
        if perm is not None:
            try:
                perm_int = int(perm, 8)
            except ValueError:
                raise ValueError('Dependency permission is incorrect')
            if (perm_int & int('600', 8)) != int('600', 8):
                raise ValueError('The owner should have rw permission')

    @classmethod
    def _check_mandatory_params(cls, params, mandatory):
        """
        Ensure that we have all mandatory parameters in the given parameters

        :param params: file parameters
        :param mandatory: required parameters
        :raises ValueError: if some of the required parameters are missing
        """
        for md in mandatory:
            if md not in params:
                raise ValueError('Mandatory parameter is missing'
                                 ': {0}'.format(md))

    def invocation_flow(self, sreq):
        """
        Invoke the backend protocol with gateway

        :param sreq: StorletRequest instance
        :return: StorletResponse instance
        """
        run_time_sbox = self.sandbox(self.scope, self.conf, self.logger)
        container_updated = self.update_container_from_cache(sreq)
        run_time_sbox.activate_storlet_daemon(sreq, container_updated)
        self._add_system_params(sreq)

        slog_path = self.paths.get_host_slog_path(sreq.storlet_main)
        storlet_pipe_path = \
            self.paths.get_host_storlet_pipe(sreq.storlet_main)

        sprotocol = StorletInvocationProtocol(sreq,
                                              storlet_pipe_path,
                                              slog_path,
                                              self.storlet_timeout,
                                              self.logger)

        sresp = sprotocol.communicate()

        self._upload_storlet_logs(slog_path, sreq)

        return sresp

    def _add_system_params(self, sreq):
        """
        Adds Storlet engine specific parameters to the invocation

        currently, this consists only of the execution path of the
        Storlet within the container.

        :params params: Request parameters
        """
        sreq.params['storlet_execution_path'] = self. \
            paths.get_sbox_storlet_dir(sreq.storlet_main)

    def _upload_storlet_logs(self, slog_path, sreq):
        """
        Upload storlet execution log as a swift object

        :param slog_path: target path
        :params sreq: ContainerStorletRequest instance
        """
        if sreq.generate_log:
            with open(slog_path, 'rb') as logfile:
                storlet_name = sreq.storlet_id.split('-')[0]
                log_obj_name = '%s.log' % storlet_name
                sreq.file_manager.put_log(log_obj_name, logfile)

    def bring_from_cache(self, obj_name, sreq, is_storlet):
        """
        Auxiliary function that:

        (1) Brings from Swift obj_name, either this is in a
            storlet or a storlet dependency.
        (2) Copies from local cache into the conrainer
        If this is a Storlet then also validates that the cache is updated
        with most recent copy of the Storlet compared to the copy residing in
        Swift.

        :params obj_name: name of the object
        :params sreq: ContainerStorletRequest instance
        :params is_storlet: True if the object is a storlet object
                            False if the object is a dependency object
        :returns: Whether the container was updated with obj_name
        """
        # Determine the cache we are to work with
        # e.g. dependency or storlet
        if is_storlet:
            cache_dir = self.paths.host_storlet_cache_dir
            get_func = sreq.file_manager.get_storlet
        else:
            cache_dir = self.paths.host_dependency_cache_dir
            get_func = sreq.file_manager.get_dependency

        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir, 0o700)

        # cache_target_path is the actual object we need to deal with
        # e.g. a concrete storlet or dependency we need to bring/update
        cache_target_path = os.path.join(cache_dir, obj_name)

        # Determine if we need to update the cache for cache_target_path
        # We default for no
        update_cache = False

        # If it does not exist in cache, we obviously need to bring
        if not os.path.isfile(cache_target_path):
            update_cache = True
        elif is_storlet:
            # The cache_target_path exists, we test if it is up-to-date
            # with the metadata we got.
            # We mention that this is currently applicable for storlets
            # only, and not for dependencies.
            # This will change when we will head dependencies as well
            fstat = os.stat(cache_target_path)
            storlet_or_size = int(
                sreq.options['storlet_content_length'].rstrip("L"))
            storlet_or_time = float(sreq.options['storlet_x_timestamp'])
            b_storlet_size_changed = fstat.st_size != storlet_or_size
            b_storlet_file_updated = float(fstat.st_mtime) < storlet_or_time
            if b_storlet_size_changed or b_storlet_file_updated:
                update_cache = True

        if update_cache:
            # If the cache needs to be updated, then get on with it
            # bring the object from storage
            data_iter, perm = get_func(obj_name)

            if perm:
                perm = int(perm, 8) & 0o700
            else:
                perm = 0o600

            # TODO(takashi): Do not directly write to target path
            with open(cache_target_path, 'wb') as fn:
                os.chmod(cache_target_path, perm)
                for data in data_iter:
                    fn.write(data)

        # The node's local cache is now updated.
        # We now verify if we need to update the
        # container itself.
        # The container needs to be updated if:
        # 1. The container does not hold a copy of the object
        # 2. The container holds an older version of the object
        update_container = False
        container_storlet_path = \
            self.paths.get_host_storlet_dir(sreq.storlet_main)
        container_target_path = os.path.join(container_storlet_path, obj_name)

        if not os.path.exists(container_storlet_path):
            os.makedirs(container_storlet_path, 0o700)
            update_container = True
        elif not os.path.isfile(container_target_path):
            update_container = True
        else:
            fstat_cached = os.stat(cache_target_path)
            fstat_container = os.stat(container_target_path)
            if fstat_cached.st_size != fstat_container.st_size:
                update_container = True
            if fstat_cached.st_mtime < fstat_container.st_mtime:
                update_container = True

        if update_container:
            # need to copy from cache to container
            # copy2 also copies the permissions
            shutil.copy2(cache_target_path, container_target_path)

        return update_container

    def update_container_from_cache(self, sreq):
        """
        Iterates over the storlet name and its dependencies appearing

        in the invocation data and make sure they are brought to the
        local cache, and from there to the container.
        Uses the bring_from_cache auxiliary function.

        :params sreq: ContainerStorletRequest instance
        :returns: True if the container was updated
        """
        # where at the host side, reside the storlet containers
        storlet_path = self.paths.host_storlet_base_dir
        if not os.path.exists(storlet_path):
            os.makedirs(storlet_path, 0o755)

        # Iterate over storlet and dependencies, and make sure
        # they are updated within the container.
        # return True if any of them wea actually
        # updated within the container
        container_updated = False

        updated = self.bring_from_cache(sreq.storlet_id, sreq, True)
        container_updated = container_updated or updated

        for dep in sreq.dependencies:
            updated = self.bring_from_cache(dep, sreq, False)
            container_updated = container_updated or updated

        return container_updated



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestHalfIdentityStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = ''
        self.additional_headers = {}
        headers = {'X-Object-Meta-Testkey': 'tester'}
        main_class = 'org.openstack.storlet.half.HalfStorlet'
        super(TestHalfIdentityStorlet, self).setUp('HalfStorlet',
                                                   'halfstorlet-1.0.jar',
                                                   main_class,
                                                   'source.txt',
                                                   headers=headers)

    def invoke_storlet(self, op, params=None, global_params=None,
                       headers=None):
        if params is not None:
            querystring = ''
            for key in params:
                querystring += '%s=%s,' % (key, params[key])
            querystring = querystring[:-1]
        else:
            querystring = None

        req_headers = {'X-Run-Storlet': self.storlet_name}
        req_headers.update(self.additional_headers)
        if headers:
            req_headers.update(headers)

        if op == 'GET':
            # Get original object
            original_h, original_c = \
                c.get_object(self.url, self.token, self.container,
                             self.storlet_file,
                             response_dict=dict())
            # print original_headers
            file_length = int(original_h['content-length'])
            processed_h, returned_c = \
                c.get_object(self.url, self.token, self.container,
                             self.storlet_file,
                             query_string=querystring, response_dict=dict(),
                             headers=req_headers, resp_chunk_size=file_length)
            processed_c = b''
            for chunk in returned_c:
                if chunk:
                    processed_c += chunk

            self.assertEqual(original_h['X-Object-Meta-Testkey'.lower()],
                             processed_h['X-Object-Meta-Testkey'.lower()])
            return processed_c

    def test_get(self):
        res = self.invoke_storlet('GET')
        self.assertEqual(b'acegikmn', res)

    def test_get_range(self):
        res = self.invoke_storlet(
            'GET',
            headers={'X-Storlet-Range': 'bytes=5-10'})
        self.assertEqual(b'fhj', res)


class TestHalfIdentityStorletOnProxy(TestHalfIdentityStorlet):
    def setUp(self):
        super(TestHalfIdentityStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase
from storlet_samples.simple.simple import SimpleStorlet
from storlets.tools.testtools import FakeStorletFileIn, FakeStorletFileOut
from tests.unit import FakeLogger
import unittest


class TestSimpleStorlet(TestCase):
    def setUp(self):
        self.logger = FakeLogger()

    def test_simple_storlet(self):
        simple_storlet = SimpleStorlet(self.logger)
        input_string = 'abcdefghijklmonp'
        store_in = [FakeStorletFileIn(input_string, {})]
        store_out = [FakeStorletFileOut()]
        params = {}
        self.assertIsNone(simple_storlet(store_in, store_out, params))
        # SimpleStorlet sets metadata {'test': 'simple'}
        self.assertEqual({'test': 'simple'}, store_out[0]._metadata)
        self.assertEqual(input_string, store_out[0].read())
        self.assertTrue(store_in[0].closed)
        self.assertTrue(store_out[0].closed)


if __name__ == '__main__':
    unittest.main()



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import contextmanager
from storlets.gateway.common.exceptions import StorletLoggerError


class StorletLogger(object):
    def __init__(self, path):
        self.log_path = path
        self._file = None

    def open(self):
        if self._file is not None:
            raise StorletLoggerError('StorletLogger is already open')

        log_dir_path = os.path.dirname(self.log_path)
        if not os.path.exists(log_dir_path):
            os.makedirs(log_dir_path, 0o700)

        self._file = open(self.log_path, 'a')
        os.chmod(self.log_path, 0o600)

    def getfd(self):
        if self._file is None:
            raise StorletLoggerError('StorletLogger is not open')
        return self._file.fileno()

    def getsize(self):
        statinfo = os.stat(self.log_path)
        return statinfo.st_size

    def close(self):
        if self._file is None:
            raise StorletLoggerError('StorletLogger is not open')

        self._file.close()
        self._file = None

    @contextmanager
    def activate(self):
        self.open()
        try:
            yield
        finally:
            self.close()



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock
from storlets.tools.deploy_storlet import main


class TestDeployStorlet(unittest.TestCase):
    # TODO(kota_): missing test cases like:
    #              - no class found in the jar
    #              - not a class in the list
    #              - fail to open the jar file
    #                (e.g. no such a file or directory)
    def test_deploy_storlet_main_java(self):
        class MockZipFile(mock.MagicMock):
            def infolist(self):
                mock_file = mock.MagicMock()
                mock_file.filename = 'main_class.class'
                return [mock_file]

            def __enter__(self):
                return self

            def __exit__(self, *args, **kwargs):
                pass

        with mock.patch('zipfile.ZipFile', MockZipFile):
            self._test_deploy_storlet_main('java')

    def test_deploy_storlet_main_python(self):
        self._test_deploy_storlet_main('python')

    def _test_deploy_storlet_main(self, lang):
        # TODO(kota_): avoid mock for ClusterConfig and get_auth
        get_auth_func_path = 'storlets.tools.deploy_storlet.get_auth'
        deploy_storlet_path = 'storlets.tools.deploy_storlet.deploy_storlet'
        with mock.patch('storlets.tools.deploy_storlet.ClusterConfig'), \
                mock.patch(deploy_storlet_path) as mock_deploy_storlet, \
                mock.patch(get_auth_func_path) as mock_auth:

            mock_auth.return_value = ('url', 'token')
            stdins = [
                lang, 'storlet_file_path', 'main_class', ''
            ]

            class MockStdin(object):
                def readline(self):
                    return stdins.pop(0)

            with mock.patch('sys.stdin', MockStdin()):
                main(['dummy_config_path'])

        # sanity
        self.assertEqual(1, mock_deploy_storlet.call_count)
        self.assertEqual(
            mock.call('url', 'token', storlet='storlet_file_path',
                      storlet_main_class='main_class', dependencies=[],
                      language=lang.title()),
            mock_deploy_storlet.call_args)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swift.common.internal_client import InternalClient
from swift.common.swob import HTTPBadRequest, Response, Range, \
    HTTPServiceUnavailable
from swift.common.utils import config_true_value
from urllib.parse import unquote

from storlets.gateway.common.exceptions import FileManagementError, \
    StorletRuntimeException
from storlets.gateway.common.file_manager import FileManager
from storlets.gateway.common.stob import StorletData


class NotStorletRequest(Exception):
    pass


class NotStorletExecution(NotStorletRequest):
    pass


def get_container_names(conf):
    return {'storlet': conf.get('storlet_container', 'storlet'),
            'dependency': conf.get('storlet_dependency', 'dependency'),
            'log': conf.get('storlet_logcontainer', 'storletlog')}


class SwiftFileManager(FileManager):
    """
    Utility class to manager i/o of storlet-related file from/to swift
    """

    def __init__(self, account, storlet_container, dependency_container,
                 log_container, conf_file, logger):
        """
        Construct SwiftFileManager instance

        :param account: swift account
        :param storlet_container: container to store storlet files
        :param dependency_container: container to storlet dependency files
        :param log_container: container to storlet storlet execution log
        :param logger: logger instance
        """
        super(SwiftFileManager, self).__init__()
        self.account = account
        self.storlet_container = storlet_container
        self.dependency_container = dependency_container
        self.log_container = log_container
        self.conf_file = conf_file
        self.logger = logger

    @property
    def client(self):
        # TODO(kota_): IMO, we need to make this to self._client environ to
        #              get rid of redundant instantiation
        return InternalClient(self.conf_file, 'SA', 1)

    def _get_object(self, container, obj, headers=None):
        """
        Utility function to be used to get object from swift

        :param container: container name
        :param obj: object name
        :param headers: request headers
        """
        self.logger.debug('GET object %s/%s/%s from swift' %
                          (self.account, container, obj))
        _, headers, data_iter = \
            self.client.get_object(self.account, container, obj, headers)
        return headers, data_iter

    def _put_object(self, container, obj, headers, fobj):
        """
        Utility function to be used to put object into swift

        :param container: container name
        :param obj: object name
        :param headers: request headers
        :param fobj: file object
        """
        self.logger.debug('PUT object %s/%s/%s to swift' %
                          (self.account, container, obj))
        self.client.upload_object(fobj, self.account, container, headers)

    def get_storlet(self, name):
        """
        Get storlet file from swift

        :param name: object name
        :returns: (iterator to read content, None)
        :raises FileManagementError: when it fails to get the object from swift
        """
        self.logger.debug('get storlet file %s from swift' % name)
        try:
            headers, data_iter = \
                self._get_object(self.storlet_container, name)
            return data_iter, None
        except Exception:
            self.logger.exception('Failed to get storlet file '
                                  '%s/%s/%s from swift' %
                                  (self.account, self.storlet_container,
                                   name))
            raise FileManagementError('Failed to get storlet file: %s' % name)

    def get_dependency(self, name):
        """
        Get dependency file from swift

        :param name: object name
        :returns: (iterator to read content, permission string)
        :raises FileManagementError: when it fails to get the object from swift
        """
        self.logger.debug('get dependency file %s from swift' % name)
        try:
            headers, data_iter = \
                self._get_object(self.dependency_container, name)
            perm = headers.get(
                'X-Object-Meta-Storlet-Dependency-Permissions')
            return data_iter, perm
        except Exception:
            self.logger.exception('Failed to get dependency file '
                                  '%s/%s/%s from swift' %
                                  (self.account, self.dependency_container,
                                   name))
            raise FileManagementError('Failed to get dependency file: %s' %
                                      name)

    def put_log(self, name, fobj):
        """
        Save storlet execution log to swift

        :param name: object name
        :param fobj: file object for log file
        :raises FileManagementError: when it fails to put the object to swift
        """
        self.logger.debug('save log file %s into swift' % name)
        try:
            headers = {'CONTENTTYPE': 'text/plain'}
            self._put_object(self.log_container, name, headers, fobj)
        except Exception:
            self.logger.exception('Failed to put log file %s/%s/%s to swift' %
                                  (self.account, self.log_container, name))
            raise FileManagementError('Failed to put log file: %s' % name)


def _request_instance_property():
    """
    Set and retrieve the request instance.
    This works to force to tie the consistency between the request path and
    self.vars (i.e. api_version, account, container, obj) even if unexpectedly
    (separately) assigned.
    """
    def getter(self):
        return self._request

    def setter(self, request):
        self._request = request
        try:
            self._extract_vaco()
        except ValueError:
            raise NotStorletRequest()

    return property(getter, setter,
                    doc="Force to tie the request to acc/con/obj vars")


class StorletBaseHandler(object):
    """
    This is an abstract handler for Proxy/Object Server middleware
    """
    request = _request_instance_property()

    def __init__(self, request, conf, gateway_conf, app, logger):
        """
        :param request: swob.Request instance
        :param conf: middleware conf dict
        :param gateway_conf: gateway conf dict
        :param app: wsgi Application
        :param logger: logger instance
        """
        self.reseller_prefix = conf.get('reseller_prefix', 'AUTH')
        self.request = request
        self.app = app
        self.logger = logger
        self.conf = conf
        self.gateway_conf = gateway_conf
        self.gateway_class = self.conf['gateway_module']
        self.sreq_class = self.gateway_class.request_class
        self.storlet_execute_on_proxy = \
            config_true_value(conf.get('storlet_execute_on_proxy', 'false'))
        containers = get_container_names(conf)
        self.storlet_container = containers['storlet']
        self.storlet_dependency = containers['dependency']
        self.log_container = containers['log']
        self.ic_conf_path = conf.get('internal_client_conf_path',
                                     '/etc/swift/internal-client.conf')

    def _setup_gateway(self):
        """
        Setup gateway instance

        """
        self.gateway = self.gateway_class(
            self.gateway_conf, self.logger, self.scope)
        self._update_storlet_parameters_from_headers()

    def _extract_vaco(self):
        """
        Set version, account, container, obj vars from self._parse_vaco result

        :raises ValueError: if self._parse_vaco raises ValueError while
                            parsing, this method doesn't care and raise it to
                            upper caller.
        """
        self._api_version, self._account, self._container, self._obj = \
            self._parse_vaco()

    @property
    def scope(self):
        """
        Scope parameter to be passed to gateway module

        In swift-storlet usecase, we create one scope per one account.
        NOTE: scope name does not include reseller prefix
        NOTE: scope name is cut to be no longer than 12 chars
        """
        if self._account.startswith(self.reseller_prefix + '_'):
            # If the account starts with reseller prefix, remove that prefix
            # from scope
            start = len(self.reseller_prefix) + 1
        else:
            start = 0
        # TODO(takashi): Using shorter name may cause scope conflicts
        end = min(start + 13, len(self.account))
        return self._account[start:end]

    @property
    def api_version(self):
        return self._api_version

    @property
    def path(self):
        return self.request.path

    @property
    def account(self):
        return self._account

    @property
    def container(self):
        return self._container

    @property
    def obj(self):
        return self._obj

    def _parse_vaco(self):
        """
        Parse method of path from self.request which depends on child class
        (Proxy or Object)

        :return: a string tuple of (version, account, container, object)
        """
        raise NotImplementedError()

    def handle_request(self):
        """
        Run storlet

        """
        raise NotImplementedError()

    @property
    def is_storlet_execution(self):
        """
        Check if the request requires storlet execution

        :return: Whether storlet should be executed
        """
        return ('X-Run-Storlet' in self.request.headers and
                self.obj)

    @property
    def is_range_request(self):
        """
        Determines whether the request is a byte-range request

        :return: Whether the request is a byte-range request
        """
        return 'Range' in self.request.headers

    @property
    def is_storlet_range_request(self):
        return 'X-Storlet-Range' in self.request.headers

    @property
    def is_storlet_multiple_range_request(self):
        if not self.is_storlet_range_request:
            return False

        r = self.request.headers['X-Storlet-Range']
        return len(Range(r).ranges) > 1

    @property
    def has_run_on_proxy_header(self):
        """
        Check whether there is a header mandating storlet execution on proxy

        :return: Whether a header exists mandating storlet execution on proxy
        """
        if 'X-Storlet-Run-On-Proxy' in self.request.headers:
            if self.request.headers['X-Storlet-Run-On-Proxy'].strip():
                raise HTTPBadRequest('X-Storlet-Run-On-Proxy header should '
                                     'be empty', request=self.request)
            return True
        return False

    @property
    def has_extra_resources_header(self):
        """
        Check whether the client specifies multi input storlet execution.

        If the request contains 'X-Storlet-Extra-Resources' header,
        Storlets gathers multiple object data into the place, which
        currently should happen in proxy-server.

        :return: Whether the request contains X-Storlet-Extra-Resources header
        """
        return 'X-Storlet-Extra-Resources' in self.request.headers

    @property
    def execute_on_proxy(self):
        return (self.has_run_on_proxy_header or
                self.has_extra_resources_header or
                self.storlet_execute_on_proxy)

    @property
    def execute_range_on_proxy(self):
        return (self.is_storlet_multiple_range_request or
                (self.is_storlet_range_request and self.execute_on_proxy))

    def is_slo_response(self, resp):
        """
        Determines whether the response is a slo one

        :param resp: swob.Response instance
        :return: Whenther the response is a slo one
        """
        self.logger.debug(
            'Verify if {0} is an SLO assembly object'.format(self.path))

        is_slo = 'X-Static-Large-Object' in resp.headers
        if is_slo:
            self.logger.debug(
                '{0} is indeed an SLO assembly object'.format(self.path))
        else:
            self.logger.debug(
                '{0} is NOT an SLO assembly object'.format(self.path))
        return is_slo

    def _update_storlet_parameters_from_headers(self):
        """
        Extract parameters for header (an alternative to parameters through
        the query string)

        """
        parameters = {}
        for param in self.request.headers:
            if param.lower().startswith('x-storlet-parameter'):
                keyvalue = self.request.headers[param]
                keyvalue = unquote(keyvalue)
                key, value = keyvalue.split(':', 1)
                parameters[key] = value
        self.request.params.update(parameters)

    def _set_metadata_in_headers(self, headers, user_metadata):
        # _set_metadata_in_headers is for user metadata
        # set by the storlet invocation. This metadata
        # should be prefixed by 'X-Object-Meta' before
        # we return things to Swift.
        # Do not call this method with metadata headers
        # coming from swift request or response.
        if user_metadata:
            for key, val in user_metadata.items():
                # TODO(kota_): we may need to discuss what type of special
                # system metadata can be overwritten in the HTTP header and
                # should encapsulate in StorletRequest/Response class
                if key.lower() == 'content-type':
                    # If and only if it's content-type, storlet app is able
                    # to overwrite the header due to the application
                    headers[key] = val
                else:
                    headers['X-Object-Meta-%s' % key] = val

    def _call_gateway(self, resp):
        """
        Call gateway module to get result of storlet execution
        in GET flow

        :param resp: swob.Response instance
        """
        raise NotImplementedError()

    def apply_storlet(self, resp):
        """
        Apply storlet on response

        :param resp: swob.Response instance
        :return: processed response
        """
        try:
            sresp = self._call_gateway(resp)

            new_headers = resp.headers.copy()

            if 'Content-Length' in new_headers:
                new_headers.pop('Content-Length')
            if 'Transfer-Encoding' in new_headers:
                new_headers.pop('Transfer-Encoding')

            # Range response(206) should be replaced by 200
            # If the range is being processed on the object node
            # then we will get 200 as the response will not have a
            # range iter.
            if 'Content-Range' in resp.headers:
                new_headers['Storlet-Input-Range'] = \
                    resp.headers['Content-Range']
                new_headers.pop('Content-Range')

            self._set_metadata_in_headers(new_headers,
                                          sresp.data.user_metadata)
            response = Response(headers=new_headers,
                                app_iter=sresp.data.data_iter,
                                request=self.request)
        except StorletRuntimeException:
            response = HTTPServiceUnavailable()

        return response

    def _get_user_metadata(self, headers):
        metadata = {}
        for key in headers:
            if key.startswith('X-Object-Meta-Storlet-'):
                pass
            elif key.startswith('X-Object-Meta-'):
                short_key = key[len('X-Object-Meta-'):]
                metadata[short_key] = headers[key]
        return metadata

    def _get_storlet_invocation_options(self, req):
        options = dict()

        filtered_key = ['X-Storlet-Range', 'X-Storlet-Generate-Log']

        for key in req.headers:
            prefix = 'X-Storlet-'
            if key.startswith(prefix) and key not in filtered_key:
                new_key = 'storlet_' + \
                    key[len(prefix):].lower().replace('-', '_')
                options[new_key] = req.headers.get(key)

        scope = self.account
        if scope.rfind(':') > 0:
            scope = scope[:scope.rfind(':')]
        options['scope'] = scope

        options['generate_log'] = \
            config_true_value(req.headers.get('X-Storlet-Generate-Log'))

        options['file_manager'] = \
            SwiftFileManager(self.account, self.storlet_container,
                             self.storlet_dependency, self.log_container,
                             self.ic_conf_path, self.logger)

        return options

    def _build_storlet_request(self, req, sheaders, sbody_iter):
        """
        Build a storlet_gatway.common.stob.StorletRequest (or its child)
        instance for storlet invocation

        :param req: an instance of swift.common.swob.Request
        :param sheaders: swift.common.swob.HeaderKeyDict instance which
                         includes object metadata information to be passed
                         to storlet daemon
        :param sbody_iter: an iterator instance to pass to storlet daemon as
                           input stream which can have _fp to suggest fd to
                           read directly.
        :return: storlet_gatway.common.stob.StorletRequest instance (or its
                 child instance)
        """
        storlet_id = req.headers.get('X-Run-Storlet')
        user_metadata = self._get_user_metadata(sheaders)
        options = self._get_storlet_invocation_options(req)

        if hasattr(sbody_iter, '_fp'):
            data = StorletData(user_metadata, data_fd=sbody_iter._fp.fileno())
        else:
            data = StorletData(user_metadata, data_iter=sbody_iter)

        return self.sreq_class(storlet_id, req.params, data, options=options)



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest

# Below is the records.txt file we are testing with
# position line content
# 0        header line
# 12       random text 1
# 26       random text 2
# 40       random text 3
# 54       random text 4
# 68       random text 5
# 82       random text 6
# 96       random text 7
# 110      random text 8
# 124      random text 9
# 138      random text 10
# 153      random text 100
# 169      random text 1000
# 186      random text 10000
# 204      ....
records_txt = \
    b"header line\n\
random text 1\n\
random text 2\n\
random text 3\n\
random text 4\n\
random text 5\n\
random text 6\n\
random text 7\n\
random text 8\n\
random text 9\n\
random text 10\n\
random text 100\n\
random text 1000\n\
random text 10000\n\
random text 1\n"


class TestPartitionsIdentityStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = 'partitionsidentitystorlet-1.0.log'
        self.additional_headers = {}
        main_class = 'org.openstack.storlet.PartitionsIdentityStorlet'
        super(TestPartitionsIdentityStorlet, self).setUp(
            'PartitionsIdentityStorlet',
            'partitionsidentitystorlet-1.0.jar',
            main_class,
            'records.txt')

    def invoke_storlet(self, start, end, first_partition, max_record_line):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        headers['X-Storlet-Range'] = 'bytes=%d-%d' % (start,
                                                      end + max_record_line)
        headers['X-Storlet-Parameter-1'] = '%s:%s' % ('start', start)
        headers['X-Storlet-Parameter-2'] = '%s:%s' % ('end', end)
        headers['X-Storlet-Parameter-3'] = '%s:%s' % ('max_record_line',
                                                      max_record_line)
        headers['X-Storlet-Parameter-4'] = '%s:%s' % ('first_partition',
                                                      first_partition)
        _, content = c.get_object(
            self.url, self.token, self.container, self.storlet_file,
            response_dict=dict(), headers=headers)
        return content

    # The input file in our case has the following characteristics:
    # It has a header line and actual data starts at byte 13
    # Since the first byte is at location 0, the 13th byte is at location 12
    # The max line is 18 chars (including '\n')
    # Here are the tests we do:
    # 1. First range tests:
    #   - range Starts with 13
    #   - firstLine param is set to True
    #   - use various max_line_len
    #   - use various end bytes
    # 2. Second range tests
    #   - range can start at any location
    #   - firstLine param is set to False
    #   - use various start, end bytes

    def _test_first_range(self, max_record_line):
        # According to the file's content end=82 should give us
        # lines 1-5, plus line 6 as the extra line
        content = self.invoke_storlet(12, 82, 'true', max_record_line)
        # byte 95 is the newline of line 6
        self.assertEqual(content, records_txt[12:95] + b'\n')
        # end=83,85,...,94 gets us deeper into the same line,
        # output should be the same:
        # lines 1-5, plus line 6 as the extra line
        for i in range(12):
            content = self.invoke_storlet(12, i + 83, 'true', max_record_line)
            self.assertEqual(content, records_txt[12:95] + b'\n')
        # Now that we move one extra character we get a different answer
        content = self.invoke_storlet(12, 95, 'true', max_record_line)
        self.assertNotEqual(content, records_txt[12:95] + b'\n')

    def _test_second_range(self, max_record_line):
        # According to the file's content start=82 should give us
        # line 7 as the first line (dropping line 6)
        # end=185 should give us line 13 as well as an extra line
        # and it should skip line 6 being the first line
        content = self.invoke_storlet(82, 185, 'false', max_record_line)
        self.assertEqual(content, records_txt[96:203] + b'\n')
        # starting from 83 up to 95 should give the exact result
        for i in range(12):
            content = self.invoke_storlet(i + 83, 185, 'false',
                                          max_record_line)
            self.assertEqual(content, records_txt[96:203] + b'\n')
        # ending at any point up to 202 should give the exact result
        for i in range(16):
            content = self.invoke_storlet(82, 186 + i, 'false',
                                          max_record_line)
            self.assertEqual(content, records_txt[96:203] + b'\n')
        # now for the combinations of the two:
        for i in range(12):
            for j in range(16):
                content = self.invoke_storlet(83 + i, 186 + j, 'false',
                                              max_record_line)
                self.assertEqual(content, records_txt[96:203] + b'\n')

    def test_second_range(self):
        self._test_second_range(80)

    def test_first_range(self):
        for max_record_line in range(5):
            # The maximum line length in this section is 14
            # includuing the new line. and so any max length >=14
            # should be good.
            self._test_first_range(80)


class TestPartitionsIdentityStorletOnProxy(TestPartitionsIdentityStorlet):
    def setUp(self):
        super(TestPartitionsIdentityStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



'''----------------------------------------------------------------
Copyright (c) 2010-2016 OpenStack Foundation

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
Limitations under the License.
----------------------------------------------------------------'''

from swiftclient import client as swift_client
from tests.functional import StorletBaseFunctionalTest
import unittest


class TestCapabilities(StorletBaseFunctionalTest):
    def setUp(self):
        super(TestCapabilities, self).setUp()

    def test_get_capabilities(self):
        _os_options = {
            'project_name': self.conf.project_name
        }
        conn = swift_client.Connection(self.conf.auth_uri,
                                       self.conf.admin_user,
                                       self.conf.admin_password,
                                       insecure=True,
                                       os_options=_os_options,
                                       auth_version=self.conf.auth_version)
        info = conn.get_capabilities()
        self.assertIn('storlet_handler', info)
        options = info['storlet_handler']
        # TODO(eranr): take values from conf
        self.assertEqual('dependency', options['storlet_dependency'])
        self.assertEqual('storlet', options['storlet_container'])
        self.assertEqual('DockerStorletGateway',
                         options['storlet_gateway_class'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
from io import BytesIO, StringIO
import json
from shutil import rmtree
from tempfile import mkdtemp
import unittest
from unittest import mock

from swift.common.swob import Response
from swift.common.utils import FileLikeIter

from storlets.sbus.client import SBusResponse
from storlets.gateway.common.stob import StorletData
from storlets.gateway.gateways.container.gateway import ContainerStorletRequest
from tests.unit import FakeLogger
from tests.unit.gateway.gateways import FakeFileManager


class TestContainerStorletRequest(unittest.TestCase):

    def test_init(self):
        # Java
        storlet_id = 'Storlet-1.0.jar'
        params = {'Param1': 'Value1', 'Param2': 'Value2'}
        metadata = {'MetaKey1': 'MetaValue1', 'MetaKey2': 'MetaValue2'}
        data = StorletData(metadata, iter(StringIO()))

        # with dependencies
        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertEqual(metadata, dsreq.data.user_metadata)
        self.assertEqual(params, dsreq.params)
        self.assertEqual('Storlet-1.0.jar', dsreq.storlet_id)
        self.assertEqual('org.openstack.storlet.Storlet', dsreq.storlet_main)
        self.assertEqual(['dep1', 'dep2'], dsreq.dependencies)
        self.assertEqual('java', dsreq.storlet_language)
        self.assertIsNone(dsreq.storlet_language_version)

        # without dependencies
        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertEqual(metadata, dsreq.data.user_metadata)
        self.assertEqual(params, dsreq.params)
        self.assertEqual('Storlet-1.0.jar', dsreq.storlet_id)
        self.assertEqual('org.openstack.storlet.Storlet', dsreq.storlet_main)
        self.assertEqual([], dsreq.dependencies)
        self.assertEqual('java', dsreq.storlet_language)
        self.assertIsNone(dsreq.storlet_language_version)

        # storlet_language is not given
        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        with self.assertRaises(ValueError):
            ContainerStorletRequest(storlet_id, params, data, options=options)

        # storlet_main is not given
        options = {'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        with self.assertRaises(ValueError):
            ContainerStorletRequest(storlet_id, params, data, options=options)

        # file_manager is not given
        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_language': 'java'}
        with self.assertRaises(ValueError):
            ContainerStorletRequest(storlet_id, params, data, options=options)

        # Python
        storlet_id = 'storlet.py'
        params = {'Param1': 'Value1', 'Param2': 'Value2'}
        metadata = {'MetaKey1': 'MetaValue1', 'MetaKey2': 'MetaValue2'}
        data = StorletData(metadata, iter(StringIO()))

        # without language version
        options = {'storlet_main': 'storlet.Storlet',
                   'storlet_language': 'python',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertEqual(metadata, dsreq.data.user_metadata)
        self.assertEqual(params, dsreq.params)
        self.assertEqual('storlet.py', dsreq.storlet_id)
        self.assertEqual('storlet.Storlet', dsreq.storlet_main)
        self.assertEqual([], dsreq.dependencies)
        self.assertEqual('python', dsreq.storlet_language)
        self.assertIsNone(dsreq.storlet_language_version)

        # with language version
        options = {'storlet_main': 'storlet.Storlet',
                   'storlet_language': 'python',
                   'storlet_language_version': '3.6',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertEqual(metadata, dsreq.data.user_metadata)
        self.assertEqual(params, dsreq.params)
        self.assertEqual('storlet.py', dsreq.storlet_id)
        self.assertEqual('storlet.Storlet', dsreq.storlet_main)
        self.assertEqual([], dsreq.dependencies)
        self.assertEqual('python', dsreq.storlet_language)
        self.assertEqual('3.6', dsreq.storlet_language_version)

    def test_init_with_range(self):
        storlet_id = 'Storlet-1.0.jar'
        params = {}
        metadata = {}
        data = StorletData(metadata, None, 0)

        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep'),
                   'range_start': 1,
                   'range_end': 6}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)

        self.assertEqual('Storlet-1.0.jar', dsreq.storlet_id)
        self.assertEqual('org.openstack.storlet.Storlet', dsreq.storlet_main)
        self.assertEqual(['dep1', 'dep2'], dsreq.dependencies)
        self.assertEqual('java', dsreq.storlet_language)
        self.assertIsNone(dsreq.storlet_language_version)
        self.assertEqual(1, dsreq.start)
        self.assertEqual(6, dsreq.end)

        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep'),
                   'range_start': 0,
                   'range_end': 0}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)

        self.assertEqual('Storlet-1.0.jar', dsreq.storlet_id)
        self.assertEqual('org.openstack.storlet.Storlet', dsreq.storlet_main)
        self.assertEqual(['dep1', 'dep2'], dsreq.dependencies)
        self.assertEqual('java', dsreq.storlet_language)
        self.assertIsNone(dsreq.storlet_language_version)
        self.assertEqual(0, dsreq.start)
        self.assertEqual(0, dsreq.end)

    def test_has_range(self):
        storlet_id = 'Storlet-1.0.jar'
        params = {}
        metadata = {}
        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep')}
        data = StorletData(metadata, None, 0)
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertFalse(dsreq.has_range)

        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep'),
                   'range_start': 1,
                   'range_end': 6}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertTrue(dsreq.has_range)

        options = {'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep'),
                   'range_start': 0,
                   'range_end': 6}
        dsreq = ContainerStorletRequest(storlet_id, params, data,
                                        options=options)
        self.assertTrue(dsreq.has_range)


class ContainerGatewayTestMixin(object):

    def setUp(self):
        # TODO(takashi): take these values from config file
        self.tempdir = mkdtemp()
        self.sconf = {
            'host_root': self.tempdir,
            'swift_dir': self.tempdir,
            'storlet_timeout': '9',
            'storlet_container': 'storlet',
            'storlet_dependency': 'dependency',
            'reseller_prefix': 'AUTH'
        }
        self.logger = FakeLogger()

        self.storlet_container = self.sconf['storlet_container']
        self.storlet_dependency = self.sconf['storlet_dependency']

        self.version = 'v1'
        self.account = 'AUTH_account'
        self.container = 'container'
        self.obj = 'object'
        self.sobj = 'storlet-1.0.jar'

        self.gateway = self.gateway_class(
            self.sconf, self.logger, self.account)

    def tearDown(self):
        rmtree(self.tempdir)

    @property
    def req_path(self):
        return self._create_proxy_path(
            self.version, self.account, self.container,
            self.obj)

    @property
    def storlet_path(self):
        return self._create_proxy_path(
            self.version, self.account, self.storlet_container,
            self.sobj)

    def _create_proxy_path(self, version, account, container, obj):
        return '/'.join(['', version, account, container, obj])

    def test_check_mandatory_params(self):
        params = {'keyA': 'valueA',
                  'keyB': 'valueB',
                  'keyC': 'valueC'}

        # all mandatory headers are included
        self.gateway_class._check_mandatory_params(
            params, ['keyA', 'keyB'])

        # some of mandatory headers are missing
        with self.assertRaises(ValueError):
            self.gateway_class._check_mandatory_params(
                params, ['keyA', 'KeyD'])

    def test_validate_storlet_registration_java(self):
        # correct name and headers w/ dependency
        obj = 'storlet-1.0.jar'
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        self.gateway_class.validate_storlet_registration(params, obj)

        # correct name and headers w/o dependency
        obj = 'storlet-1.0.jar'
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        self.gateway_class.validate_storlet_registration(params, obj)

        # some header keys are missing
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # wrong name
        obj = 'storlet.jar'
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

    def test_validate_storlet_registration_python(self):
        # correct name and headers w/ dependency
        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Language-Version': '3.6',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet.Storlet'}
        self.gateway_class.validate_storlet_registration(params, obj)

        # wrong version
        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Language-Version': '1.7',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet.Storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # py2 is no more supported
        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Language-Version': '2.7',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet.Storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # wrong name
        obj = 'storlet.pyfoo'
        params = {'Language': 'python',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet.Storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # wrong main class
        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'another_storlet.Storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        obj = 'storlet.py'
        params = {'Language': 'python',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'storlet.foo.Storlet'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

    def test_validate_storlet_registration_not_suppoeted(self):
        # unsupported language
        obj = 'storlet.foo'
        params = {'Language': 'bar',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # same name for storlet and dependency
        obj = 'storlet-1.0.jar'
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Dependency': 'storlet-1.0.jar',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

        # duplicated name in dependencies
        obj = 'storlet-1.0.jar'
        params = {'Language': 'java',
                  'Interface-Version': '1.0',
                  'Dependency': 'dep_file,dep_file',
                  'Object-Metadata': 'no',
                  'Main': 'path.to.storlet.class'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_storlet_registration(params, obj)

    def test_validate_dependency_registration(self):
        # w/o dependency parameter
        obj = 'dep_file'
        params = {'Dependency-Version': '1.0'}
        self.gateway_class.validate_dependency_registration(params, obj)

        # w/ correct dependency parameter
        params = {
            'Dependency-Permissions': '755',
            'Dependency-Version': '1.0'}
        self.gateway_class.validate_dependency_registration(params, obj)

        # w/ wrong dependency parameter
        params = {
            'Dependency-Permissions': '400',
            'Dependency-Version': '1.0'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_dependency_registration(params, obj)

        # w/ invalid dependency parameter
        params = {
            'Dependency-Permissions': 'foo',
            'Dependency-Version': '1.0'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_dependency_registration(params, obj)
        params = {
            'Dependency-Permissions': '888',
            'Dependency-Version': '1.0'}
        with self.assertRaises(ValueError):
            self.gateway_class.validate_dependency_registration(params, obj)

    def _test_invocation_flow(self, extra_sources=None):
        extra_sources = extra_sources or []
        options = {'generate_log': False,
                   'scope': 'AUTH_account',
                   'storlet_main': 'org.openstack.storlet.Storlet',
                   'storlet_dependency': 'dep1,dep2',
                   'storlet_language': 'java',
                   'file_manager': FakeFileManager('storlet', 'dep')}

        data = StorletData(
            user_metadata={},
            data_iter=iter('body'))
        st_req = ContainerStorletRequest(
            storlet_id=self.sobj,
            params={},
            data=data,
            options=options)

        # TODO(kota_): need more efficient way for emuration of return value
        # from SDaemon
        value_generator = iter([
            # first, we get metadata json
            json.dumps({'metadata': 'return'}),
            # then we get object data
            'something', '',
        ])

        def mock_read(fd, size):
            try:
                value = next(value_generator)
            except StopIteration:
                raise Exception('called more then expected')
            # NOTE(takashi): Make sure that we return bytes in PY3
            return value.encode('utf-8')

        def mock_close(fd):
            pass

        called_fd_and_bodies = []
        invocation_protocol = \
            'storlets.gateway.gateways.container.runtime.' \
            'StorletInvocationProtocol._write_input_data'

        def mock_writer(self, fd, app_iter):
            body = ''
            for chunk in app_iter:
                body += chunk
            called_fd_and_bodies.append((fd, body))

        # prepare nested mock patch
        # SBus -> mock SBus.send() for container communication
        # os.read -> mock reading the file descriptor from container
        # select.select -> mock fd communication which can be readable
        @mock.patch('storlets.gateway.gateways.container.runtime.SBusClient')
        @mock.patch('storlets.gateway.gateways.container.runtime.os.read',
                    mock_read)
        @mock.patch('storlets.gateway.gateways.container.runtime.os.close',
                    mock_close)
        @mock.patch('storlets.gateway.gateways.container.runtime.select.'
                    'select',
                    lambda r, w, x, timeout=None: (r, w, x))
        @mock.patch('storlets.gateway.common.stob.os.read', mock_read)
        @mock.patch(invocation_protocol, mock_writer)
        def test_invocation_flow(client):
            client.ping.return_value = SBusResponse(True, 'OK')
            client.stop_daemon.return_value = SBusResponse(True, 'OK')
            client.start_daemon.return_value = SBusResponse(True, 'OK')
            client.execute.return_value = SBusResponse(True, 'OK', 'someid')

            sresp = self.gateway.invocation_flow(st_req)
            eventlet.sleep(0.1)
            file_like = FileLikeIter(sresp.data.data_iter)
            self.assertEqual(b'something', file_like.read())

        # I hate the decorator to return an instance but to track current
        # implementation, we have to make a mock class for this. Need to fix.

        class MockFileManager(object):
            def get_storlet(self, req):
                return BytesIO(b'mock'), None

            def get_dependency(self, req):
                return BytesIO(b'mock'), None

        st_req.file_manager = MockFileManager()
        st_req.extra_data_list = extra_sources

        test_invocation_flow()

        # ensure st_req.app_iter is drawn
        self.assertRaises(StopIteration, next, st_req.data.data_iter)
        expected_mock_writer_calls = len(extra_sources) + 1
        self.assertEqual(expected_mock_writer_calls,
                         len(called_fd_and_bodies))
        self.assertEqual('body', called_fd_and_bodies[0][1])
        return called_fd_and_bodies

    def test_invocation_flow(self):
        self._test_invocation_flow()

    def test_invocation_flow_with_extra_sources(self):
        data_sources = []

        def generate_extra_data():
            sw_resp = Response(
                app_iter=iter(['This is a response body']), status=200)

            data = StorletData(
                user_metadata={},
                data_iter=sw_resp.app_iter)
            data_sources.append(sw_resp.app_iter)
            return data

        extra_data = generate_extra_data()
        mock_calls = self._test_invocation_flow(extra_sources=[extra_data])
        self.assertEqual('This is a response body', mock_calls[1][1])

        # run all existing eventlet threads
        for app_iter in data_sources:
            # ensure all app_iters are drawn
            self.assertRaises(StopIteration, next, app_iter)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from swift.common.swob import Request, HTTPOk, HTTPCreated
from storlets.swift_middleware.handlers import StorletObjectHandler

from tests.unit.swift_middleware.handlers import \
    BaseTestStorletMiddleware, create_handler_config


class TestStorletMiddlewareObject(BaseTestStorletMiddleware):
    def setUp(self):
        super(TestStorletMiddlewareObject, self).setUp(exec_server='object')

    def test_load_app(self):
        try:
            self.get_app(self.base_app, self.conf)
        except Exception:
            self.fail('Application loading got an error')

    def test_call_unsupported_method(self):
        def call(method):
            path = '/sda1/p/AUTH_a/c/o'
            headers = {'X-Backend-Storage-Policy-Index': '0',
                       'X-Run-Storlet': 'Storlet-1.0.jar'}
            req = Request.blank(path, environ={'REQUEST_METHOD': method},
                                headers=headers)
            resp = self.get_response(req)
            self.assertEqual('405 Method Not Allowed', resp.status)

        for method in ('POST', 'PUT', 'DELETE'):
            call(method)

    def test_PUT_objet_into_storlets_container(self):
        target = '/sda1/p/AUTH_a/storlet/storlet-1.0.jar'
        self.base_app.register('PUT', target, HTTPCreated)

        sheaders = {'X-Object-Meta-Storlet-Language': 'Java',
                    'X-Object-Meta-Storlet-Interface-Version': '1.0',
                    'X-Object-Meta-Storlet-Dependency': 'dependency',
                    'X-Object-Meta-Storlet-Main':
                        'org.openstack.storlet.Storlet'}
        req = Request.blank(target, environ={'REQUEST_METHOD': 'PUT'},
                            headers=sheaders)
        resp = self.get_response(req)
        self.assertEqual('201 Created', resp.status)

    def test_GET_without_storlets(self):
        def basic_get(path):
            req = Request.blank(
                path, environ={'REQUEST_METHOD': 'GET'},
                headers={'X-Backend-Storage-Policy-Index': '0'})
            self.base_app.register('GET', path, HTTPOk, body=b'FAKE APP')
            resp = self.get_response(req)
            self.assertEqual('200 OK', resp.status)
            self.assertEqual(b'FAKE APP', resp.body)
            self.base_app.reset_all()

        for target in ('/sda1/p/AUTH_a', '/sda1/p/AUTH_a/c',
                       '/sda1/p/AUTH_a/c/o'):
            basic_get(target)

    def test_GET_with_storlets(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        resp = self.get_response(req)
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'FAKE APP', resp.body)

    def test_GET_with_storlets_and_http_range(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'Range': 'bytes=10-20'})
        resp = self.get_response(req)
        self.assertEqual('416 Requested Range Not Satisfiable',
                         resp.status)

    def test_GET_with_storlets_and_storlet_range(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Range': 'bytes=1-6',
                     'Range': 'bytes=1-6'})
        resp = self.get_response(req)
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'AKE AP', resp.body)
        self.assertEqual('bytes 1-6/8',
                         resp.headers.get('Storlet-Input-Range'))

    def test_GET_with_storlets_and_storlet_proxy_range(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk, body=b'FAKE APP')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Range': 'bytes=1-6',
                     'Range': 'bytes=1-6',
                     'X-Storlet-Run-On-Proxy': ''})
        resp = self.get_response(req)
        self.assertEqual('206 Partial Content', resp.status)
        self.assertEqual(b'AKE AP', resp.body)
        self.assertEqual('bytes 1-6/8',
                         resp.headers.get('Content-Range'))

    def test_GET_slo_manifest_with_storlets(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk,
                               headers={'X-Static-Large-Object': 'True'},
                               body=b'FAKE MANIFEST')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        resp = self.get_response(req)
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'FAKE MANIFEST', resp.body)

    def test_GET_slo_segment_with_storlets(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register('GET', target, HTTPOk,
                               headers={'X-Static-Large-Object': 'True'},
                               body=b'FAKE SEGMENT')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'multipart-manifest': 'get',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        resp = self.get_response(req)
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'FAKE SEGMENT', resp.body)

    def test_GET_symlink_with_storlets(self):
        target = '/sda1/p/AUTH_a/c/o'
        self.base_app.register(
            'GET', target, HTTPOk,
            headers={'X-Object-Sysmeta-Symlink-Target': 'c/o2'},
            body=b'FAKE CONTENT')

        req = Request.blank(
            target, environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        resp = self.get_response(req)
        self.assertEqual('200 OK', resp.status)
        self.assertEqual(b'FAKE CONTENT', resp.body)

    def test_storlets_with_invalid_method(self):
        target = '/sda1/p/AUTH_a/c/o'

        req = Request.blank(
            target, environ={'REQUEST_METHOD': '_parse_vaco'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        resp = self.get_response(req)
        self.assertEqual('405 Method Not Allowed', resp.status)


class TestStorletObjectHandler(unittest.TestCase):
    def setUp(self):
        self.handler_class = StorletObjectHandler
        self.conf = create_handler_config('object')
        self.gateway_conf = {}

    def test_init_handler(self):
        req = Request.blank(
            '/dev/part/acc/cont/obj', environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        handler = self.handler_class(
            req, self.conf, self.gateway_conf, mock.MagicMock(),
            mock.MagicMock())
        # FIXME: still hold api version 0 at ObjectHandler but will be
        #        deprecated if it's never used.
        self.assertEqual('0', handler.api_version)
        self.assertEqual('acc', handler.account)
        self.assertEqual('cont', handler.container)
        self.assertEqual('obj', handler.obj)

        req = Request.blank(
            '/dev/part/acc2/cont2/obj2', environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar'})
        handler.request = req
        self.assertEqual('/dev/part/acc2/cont2/obj2', handler.request.path)
        self.assertEqual('0', handler.api_version)
        self.assertEqual('acc2', handler.account)
        self.assertEqual('cont2', handler.container)
        self.assertEqual('obj2', handler.obj)

        # no direct assignment allowed
        with self.assertRaises(AttributeError):
            handler.api_version = '1'

        with self.assertRaises(AttributeError):
            handler.account = 'acc'

        with self.assertRaises(AttributeError):
            handler.container = 'cont'

        with self.assertRaises(AttributeError):
            handler.obj = 'obj'

    def test_get_storlet_invocation_options(self):
        req = Request.blank(
            '/dev/part/acc/cont/obj',
            environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Foo': 'baa'})
        handler = self.handler_class(
            req, self.conf, self.gateway_conf, mock.MagicMock(),
            mock.MagicMock())

        options = handler._get_storlet_invocation_options(req)
        self.assertEqual('baa', options['storlet_foo'])
        self.assertFalse(options['generate_log'])

        req = Request.blank(
            '/dev/part/acc/cont/obj',
            environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Foo': 'baa',
                     'X-Storlet-Generate-Log': 'True'})
        handler = self.handler_class(
            req, self.conf, self.gateway_conf, mock.MagicMock(),
            mock.MagicMock())

        options = handler._get_storlet_invocation_options(req)
        self.assertEqual('baa', options['storlet_foo'])
        self.assertTrue(options['generate_log'])

        req = Request.blank(
            '/dev/part/acc/cont/obj',
            environ={'REQUEST_METHOD': 'GET'},
            headers={'X-Backend-Storage-Policy-Index': '0',
                     'X-Run-Storlet': 'Storlet-1.0.jar',
                     'X-Storlet-Foo': 'baa',
                     'X-Storlet-Range': 'bytes=1-6',
                     'Range': 'bytes=1-6'})
        handler = self.handler_class(
            req, self.conf, self.gateway_conf, mock.MagicMock(),
            mock.MagicMock())

        options = handler._get_storlet_invocation_options(req)
        self.assertEqual('baa', options['storlet_foo'])
        self.assertFalse(options['generate_log'])
        self.assertNotIn('storlet_range', options)
        self.assertEqual(1, options['range_start'])
        self.assertEqual(7, options['range_end'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from swiftclient import client
from tests.functional.python import StorletPythonFunctionalTest
import unittest
from urllib.request import Request, urlopen


class TestSimpleStorlet(StorletPythonFunctionalTest):
    def setUp(self, version=None):
        self.storlet_log = 'simple.log'
        self.content = b'abcdefghijklmonp'
        self.additional_headers = {}
        super(TestSimpleStorlet, self).setUp(
            storlet_dir='simple',
            storlet_name='simple.py',
            storlet_main='simple.SimpleStorlet',
            storlet_file='source.txt',
            version=version)

    def test_get(self):
        resp = dict()
        req_headers = {'X-Run-Storlet': self.storlet_name}
        headers, content = client.get_object(
            self.url, self.token, self.container, self.storlet_file,
            response_dict=resp, headers=req_headers)
        self.assertEqual(200, resp['status'])
        self.assertEqual('simple', headers['x-object-meta-test'])
        self.assertEqual(self.content, content)

    def test_range_get(self):
        resp = dict()
        req_headers = {'X-Run-Storlet': self.storlet_name,
                       'X-Storlet-Range': 'bytes=1-4'}
        headers, content = client.get_object(
            self.url, self.token, self.container, self.storlet_file,
            response_dict=resp, headers=req_headers)
        self.assertEqual(200, resp['status'])
        self.assertEqual('simple', headers['x-object-meta-test'])
        self.assertEqual(self.content[1:5], content)

    def test_put(self):
        objname = self.storlet_file + '-put'

        resp = dict()
        req_headers = {'X-Run-Storlet': self.storlet_name}
        client.put_object(
            self.url, self.token, self.container, objname,
            self.content, response_dict=resp, headers=req_headers)
        self.assertEqual(201, resp['status'])

        resp = dict()
        headers, content = client.get_object(
            self.url, self.token, self.container, objname,
            response_dict=resp)
        self.assertEqual(200, resp['status'])
        self.assertEqual('simple', headers['x-object-meta-test'])
        self.assertEqual(self.content, content)

        resp = dict()
        client.delete_object(
            self.url, self.token, self.container, objname,
            response_dict=resp)
        self.assertEqual(204, resp['status'])

    def test_put_10MB_file(self):
        with tempfile.NamedTemporaryFile() as f:
            with open(f.name, 'wb') as wf:
                for _ in range(10):
                    wf.write(b'1' * (1 * 1024 * 1024))

            headers = {'X-Run-Storlet': self.storlet_name}
            headers.update(self.additional_headers)
            with open(f.name, 'rb') as rf:
                response = dict()
                client.put_object(self.url, self.token,
                                  self.container, f.name, rf,
                                  10 * 1024 * 1024, None, None,
                                  "application/octet-stream",
                                  headers, None, None, None, response)

            status = response.get('status')
            self.assertEqual(201, status)

    test_put_10MB_file.slow = 1

    def test_copy_from(self):
        resp = dict()
        objname = self.storlet_file + '-copy'
        req_headers = {'X-Run-Storlet': self.storlet_name,
                       'X-Copy-From': '%s/%s' %
                       (self.container, self.storlet_file)}
        client.put_object(
            self.url, self.token, self.container, objname,
            self.content, response_dict=resp, headers=req_headers)

        self.assertEqual(201, resp['status'])
        resp_header = resp['headers']
        self.assertEqual('%s/%s' % (self.container, self.storlet_file),
                         resp_header['x-storlet-generated-from'])
        self.assertEqual(self.acct,
                         resp_header['x-storlet-generated-from-account'])
        self.assertIn('x-storlet-generated-from-last-modified', resp_header)

        headers = client.head_object(self.url, self.token,
                                     self.container, objname)
        self.assertEqual(str(len(self.content)), headers['content-length'])

        resp = dict()
        client.delete_object(
            self.url, self.token, self.container, objname,
            response_dict=resp)
        self.assertEqual(204, resp['status'])

    def test_copy_dest(self):
        # No COPY in swiftclient. Using urllib instead...
        url = os.path.join(self.url, self.container, self.storlet_file)
        objname = self.storlet_file + '-copy-ex'
        headers = {'X-Auth-Token': self.token,
                   'X-Run-Storlet': self.storlet_name,
                   'Destination': '%s/%s' % (self.container, objname)}
        headers.update(self.additional_headers)
        req = Request(url, headers=headers)
        req.get_method = lambda: 'COPY'
        conn = urlopen(req, timeout=10)

        self.assertEqual(201, conn.getcode())
        self.assertEqual('%s/%s' % (self.container, self.storlet_file),
                         conn.info()['x-storlet-generated-from'])
        self.assertEqual(self.acct,
                         conn.info()['x-storlet-generated-from-account'])
        self.assertIn('x-storlet-generated-from-last-modified', conn.info())

        headers = client.head_object(self.url, self.token,
                                     self.container, objname)
        self.assertEqual(str(len(self.content)), headers['content-length'])

        resp = dict()
        client.delete_object(
            self.url, self.token, self.container, objname,
            response_dict=resp)
        self.assertEqual(204, resp['status'])


class TestSimpleStorletOnProxy(TestSimpleStorlet):
    def setUp(self):
        super(TestSimpleStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
import os
import tempfile
import unittest
from unittest import mock

from storlets.gateway.common.stob import FileDescriptorIterator


class TestFileDescriptorIterator(unittest.TestCase):

    def setUp(self):
        self.timeout = 10
        self.content = b'aaaa\nbbbb\ncccc\n'

        # TODO(takashi): TemporaryFile may be safer, but causes OS Error in
        #                close in IterLike.__del__
        self.fd, fname = tempfile.mkstemp()
        os.unlink(fname)
        os.write(self.fd, self.content)
        self._reset_fd()
        self.iter_like = FileDescriptorIterator(
            self.fd, self.timeout, None)

    def tearDown(self):
        pass

    def _fake_select(self, r, w, x, timeout=0):
        return r, w, x

    def _reset_fd(self):
        os.lseek(self.fd, 0, os.SEEK_SET)

    @contextmanager
    def _mock_select(self):
        # TODO(takashi): This is needed to avoid PermissionError in UT
        with mock.patch('storlets.gateway.common.stob.select.select',
                        self._fake_select):
            yield

    def test_read_with_timeout(self):
        self.assertEqual(
            b'aaaa\nb', self.iter_like.read_with_timeout(6))
        self.assertEqual(
            b'bbb\ncc', self.iter_like.read_with_timeout(6))
        self.assertEqual(
            b'cc\n', self.iter_like.read_with_timeout(6))
        self.assertEqual(
            b'', self.iter_like.read_with_timeout(6))

    def test_next(self):
        with self._mock_select():
            self.assertEqual(b'aaaa\nb', self.iter_like.next(6))
            self.assertEqual(b'bbb\ncc', self.iter_like.next(6))
            self.assertEqual(b'cc\n', self.iter_like.next(6))
            with self.assertRaises(StopIteration):
                self.iter_like.next(6)
        self._reset_fd()

        with self._mock_select():
            # if size > content length
            self.assertEqual(b'aaaa\nbbbb\ncccc\n',
                             self.iter_like.next(50))
            with self.assertRaises(StopIteration):
                self.iter_like.next(50)

    def test_read(self):
        with self._mock_select():
            self.assertEqual(b'aaaa\nb', self.iter_like.read(6))
            self.assertEqual(b'bbb\ncc', self.iter_like.read(6))
            self.assertEqual(b'cc\n', self.iter_like.read(6))
            with self.assertRaises(StopIteration):
                self.iter_like.next(6)

    def test_readline(self):
        with self._mock_select():
            # if size = -1
            self.assertEqual(b'aaaa\n', self.iter_like.readline())

            # if size < line length
            self.assertEqual(b'bb', self.iter_like.readline(2))

            # read remaining chars in line
            self.assertEqual(b'bb\n', self.iter_like.readline())

            # if size > line length
            self.assertEqual(b'cccc\n', self.iter_like.readline(100))
            with self.assertRaises(StopIteration):
                self.iter_like.readline()

    def test_readlines(self):
        with self._mock_select():
            self.assertEqual(
                [b'aaaa\n', b'bbbb\n', b'cccc\n'],
                self.iter_like.readlines())
        self._reset_fd()

        with self._mock_select():
            self.assertEqual(
                [b'aaaa\n', b'bb'],
                self.iter_like.readlines(7))


if __name__ == '__main__':
    unittest.main()
